#!/usr/bin/env python3
#

''' Support for DeDRM/noDRM.

    This is an experimental module aimed at making the DeDRM/noDRM
    packages run outside Calibre's plugin environment.
    It does not yet work.
'''

from contextlib import contextmanager, redirect_stdout
from dataclasses import dataclass, field
from datetime import datetime
from getopt import GetoptError
import importlib
import json
import os
from os.path import (
    basename,
    isdir as isdirpath,
    join as joinpath,
    normpath,
    realpath,
    splitext,
)
import sys
from tempfile import NamedTemporaryFile, TemporaryDirectory
import time
from typing import Iterable, List, Optional

from cs.cmdutils import BaseCommand
from cs.context import stackattrs
from cs.deco import fmtdoc, Promotable
from cs.fileutils import atomic_filename
from cs.lex import r, stripped_dedent
from cs.logutils import warning
from cs.pfx import Pfx, pfx_call, pfx_method
from cs.sqltags import SQLTags
from cs.upd import print  # pylint: disable=redefined-builtin

DEDRM_PACKAGE_PATH_ENVVAR = 'DEDRM_PACKAGE_PATH'

def main(argv=None):
  ''' DeDRM command line mode.
  '''
  return DeDRMCommand(argv).run()

class DeDRMCommand(BaseCommand):

  GETOPT_SPEC = 'D:'
  USAGE_FORMAT = r'''Usage: {cmd} [-D dedrm_package_path] subcommand [args...]
    -D  Specify the filesystem path to the DeDRM/noDRM plugin top level.
        For example, if you had a checkout of git@github.com:noDRM/DeDRM_tools.git
        at /path/to/DeDRM_tools--noDRM you could supply:
        -D /path/to/DeDRM_tools--noDRM/DeDRM_plugin
        or place that value in the $DEDRM_PACKAGE_PATH environment variable.
  '''

  @dataclass
  class Options(BaseCommand.Options):
    dedrm_package_path: Optional[str] = field(
        default_factory=lambda: os.environ.get(DEDRM_PACKAGE_PATH_ENVVAR)
    )

  def apply_opt(self, opt, val):
    badopt = False
    if opt == '-D':
      self.options.dedrm_package_path = val
    else:
      raise RuntimeError("unhandled option")
    if badopt:
      raise GetoptError("bad option value")

  @contextmanager
  def run_context(self):
    with super().run_context():
      options = self.options
      with Pfx("dedrm_package_path"):
        dedrm_package_path = options.dedrm_package_path
        if dedrm_package_path is None:
          if not os.environ.get(DEDRM_PACKAGE_PATH_ENVVAR):
            raise GetoptError(
                f'no ${DEDRM_PACKAGE_PATH_ENVVAR} and no -D option'
            )
        try:
          options.dedrm = pfx_call(DeDRMWrapper, dedrm_package_path)
        except ValueError as e:
          raise GetoptError("bad dedrm_package_path: %s" % (e,))
        options.dedrm_package_path = options.dedrm.dedrm_package_path
      yield

  def cmd_import(self, argv):
    ''' Usage: {cmd} module_name...
          Exercise the DeDRM python import mechanism for each module_name.
    '''
    if not argv:
      raise GetoptError("missing module_name")
    dedrm = self.options.dedrm
    xit = 0
    for module_name in argv:
      print("import", module_name, "...")
      with Pfx("import %r", module_name):
        try:
          module_name, name = module_name.split(':')
        except ValueError:
          name = None
        with dedrm.dedrm_imports():
          try:
            M = pfx_call(importlib.import_module, module_name)
          except ImportError as e:
            warning("import fails: %s", e)
            xit = 1
            continue
          print("import %r => %s" % (module_name, r(M)))
        with Pfx('.%s', name):
          if name is not None:
            try:
              value = getattr(M, name)
            except AttributeError as e:
              warning("not present: %s", e)
              xit = 1
              continue
            print("  .%r => %s" % (name, r(value)))
    return xit

  def cmd_kindlekeys(self, argv):
    ''' Usage: {cmd} [import]
          import    Read a JSON list of key dicts and update the cached keys.
    '''
    dedrm = self.options.dedrm
    if not argv:
      with redirect_stdout(sys.stderr):
        kks = dedrm.kindlekeys
      print(json.dumps(kks))
      return 0
    op = argv.pop(0)
    if op != 'import':
      raise GetoptError("expected 'import', got %r" % (op,))
    with Pfx(op):
      if argv:
        raise GetoptError("extra arguments: %r" % (argv,))
      new_kks = json.loads(sys.stdin.read())
      assert all(isinstance(kk, dict) for kk in new_kks)
      dedrm.update_kindlekeys_from_keys(new_kks)

  def cmd_remove(self, argv):
    ''' Usage: {cmd} filenames...
          Remove DRM from the specified filenames.
    '''
    dedrm = self.options.dedrm
    exists_ok = False
    output_dirpath = '.'
    # TODO: -f exists_ok -O output_dirpath
    if not argv:
      raise GetoptError("missing filenames")
    for filename in argv:
      with Pfx(filename):
        output_filename = normpath(
            joinpath(
                realpath(output_dirpath), 'decrypted-' + basename(filename)
            )
        )
        pfx_call(dedrm.remove, filename, output_filename, exists_ok=exists_ok)

class DeDRMWrapper(Promotable):
  ''' Class embodying the DeDRM/noDRM package actions.
  '''

  # The package name to use for the DeDRM/noDRM package.
  # We also use these symbols to convince the DRM package to run
  # without being installed as a calibre plugin inside calibre.
  DEDRM_PACKAGE_NAME = 'dedrm'
  DEDRM_PLUGIN_NAME = 'DeDRM'
  DEDRM_PLUGIN_VERSION = '7.2.1'

  @pfx_method
  @fmtdoc
  def __init__(
      self,
      dedrm_package_path: Optional[str] = None,
      sqltags: Optional[SQLTags] = None
  ):
    ''' Initialise the DeDRM package.

        Parameters:
        * `dedrm_package_path`: optional filesystem path to the
          DeDRM/noDRM package plugin subdirectory, default from
          ${DEDRM_PACKAGE_PATH_ENVVAR}
        * `sqltags`: optional `SQLTags` instance to store state
          under the `self.DEDRM_PACKAGE_NAME.` prefix;
          default from `SQLTags()`
    '''
    with Pfx("dedrm_package_path %r", dedrm_package_path):
      if dedrm_package_path is None:
        dedrm_package_path = os.environ.get(DEDRM_PACKAGE_PATH_ENVVAR)
        if dedrm_package_path is None:
          raise ValueError(
              f'no dedrm_package_path and no ${DEDRM_PACKAGE_PATH_ENVVAR}'
          )
        with Pfx('$%s=%r', DEDRM_PACKAGE_PATH_ENVVAR, dedrm_package_path):
          self.__init__(dedrm_package_path, sqltags=sqltags)
          return
      if not isdirpath(dedrm_package_path):
        raise ValueError("not a directory")
      if not isdirpath(joinpath(dedrm_package_path, 'standalone')):
        raise ValueError("no \"standalone\" subdirectory")
      self.dedrm_package_path = dedrm_package_path
      dedrm_DeDRM = self.import_name(self.DEDRM_PACKAGE_NAME, 'DeDRM')

      class CSEBookDeDRM(DeDRMOverride, dedrm_DeDRM):
        ''' Our wrapper for the DeDRM/noDRM `DeDRM` class
            using overrides from `DeDRMOverride`.
        '''
        alfdir = dedrm_package_path

      self.dedrm = CSEBookDeDRM()
    with self.dedrm_imports():
      kindlekey = self.import_name('kindlekey')
      # monkey patch the kindlekey.kindlekeys function
      self.base_kindlekeys = kindlekey.kindlekeys
      kindlekey.kindlekeys = self.cached_kindlekeys
      # monkey patch the kindlekey.CryptUnprotectData class
      BaseCryptUnprotectData = kindlekey.CryptUnprotectData
      LibCrypto = getLibCrypto()

      class CryptUnprotectData(BaseCryptUnprotectData):

        def __init__(self, *args):
          self.crp = LibCrypto()
          super().__init__(*args)

      kindlekey.CryptUnprotectData = CryptUnprotectData
    if sqltags is None:
      sqltags = SQLTags()
    self.tags = sqltags.subdomain(f'{self.DEDRM_PACKAGE_NAME}')

  @contextmanager
  def dedrm_imports(self):
    ''' Context manager to run some code with `self.dedrm_package_path`
        prepended to `sys.path` for import purposes.

        This also crafts some stub modules to convince the plugin
        to run by providing some things normally provided by the
        Calibre plugin environment.
    '''

    def write_module(name, contents):
      ''' Write Python code `contents` to a top level module named `name`.
      '''
      module_path = joinpath(tmpdirpath, f'{name}.py')
      with pfx_call(open, module_path, 'w') as pyf:
        print(
            stripped_dedent(
                f'''
                #!/usr/bin/env python3
                #
                # Generated by {__file__} at {datetime.now()}.
                # - Cameron Simpson <cs@cskk.id.au>
                #
                '''
            ),
            file=pyf
        )
        print(stripped_dedent(contents), file=pyf)
      ##os.system(f'set -x; cat {module_path}')

    with TemporaryDirectory(prefix='dedrm_lib') as tmpdirpath:
      # present the dedrm package as DEDRM_PACKAGE_NAME ('dedrm')
      os.symlink(
          self.dedrm_package_path,
          joinpath(tmpdirpath, self.DEDRM_PACKAGE_NAME)
      )
      # fake up __version.py
      write_module(
          '__version',
          f'''
          PLUGIN_NAME = {self.DEDRM_PLUGIN_NAME!r}
          PLUGIN_VERSION = {self.DEDRM_PLUGIN_VERSION!r}
          PLUGIN_VERSION_TUPLE = {tuple(map(int, self.DEDRM_PLUGIN_VERSION.split('.')))!r}
        ''',
      )
      # fake up prefs.DeDRM_Prefs()
      write_module(
          'prefs',
          f'''
          import os
          import {self.DEDRM_PACKAGE_NAME}.prefs
          from {self.DEDRM_PACKAGE_NAME}.prefs import DeDRM_Prefs as BaseDeDRM_Prefs
          from {self.DEDRM_PACKAGE_NAME}.standalone.jsonconfig import JSONConfig as BaseJSONConfig
          from cs.ebooks.dedrm import DeDRM_PrefsOverride
          from cs.gimmicks import warning
          class DeDRM_Prefs(DeDRM_PrefsOverride, BaseDeDRM_Prefs):
            def __init__(self, json_path=None):
              if json_path is not None:
                warning(
                    "DeDRM_Prefs: ignoring json_path=%r, using %r", json_path,
                    os.devnull
                )
              BaseDeDRM_Prefs.__init__(self, json_path=os.devnull)
          class JSONConfig(BaseJSONConfig):
            def __init__(self, rel_path_to_cf_file):
              super().__init__(rel_path_to_cf_file)
              warning(
                  "JSONConfig: replacing file_path=%r with %r",
                  self.file_path,
                  os.devnull
              )
              self.file_path = os.devnull
          {self.DEDRM_PACKAGE_NAME}.prefs.JSONConfig = JSONConfig
          ''',
      )
      with stackattrs(
          sys,
          path=[tmpdirpath, joinpath(tmpdirpath, self.DEDRM_PACKAGE_NAME)] +
          sys.path):
        import builtins
        with stackattrs(builtins, print=print):
          with redirect_stdout(sys.stderr):
            import prefs  # imported for its side effect
            yield

  def import_name(self, module_name, name=None, *, package=None):
    ''' Shim for `importlib.import_module` to import from the DeDRM/noDRM package.

        Parameters:
        * `module_name`: the module name to import
        * `name`: optional name from `module_name`
        * `package`: `package` argument for `importlib.import_module`,
          required if `module_name` is a relative import

        If `name` is omitted, return the imported module.
        Otherwise return the value of `name` from the imported module.
    '''
    with self.dedrm_imports():
      M = pfx_call(importlib.import_module, module_name, package=package)
      if name is None:
        return M
      return pfx_call(getattr, M, name)

  def remove(self, srcpath, dstpath, exists_ok=False):
    ''' Remove the DRM from `srcpath`, writing the resulting file to `dstpath`.

        Parameters:
        * `exists_ok`: if true then `dstpath` may already exist; default `False`
    '''
    with atomic_filename(dstpath, exists_ok=exists_ok) as T:
      dedrm = self.dedrm
      # monkey patch temporary_file method to return tmpfilename
      with stackattrs(dedrm, temporary_file=lambda ext: T):
        with self.dedrm_imports():
          dedrm.starttime = time.time()
          # infer book type from file extension
          booktype = splitext(basename(srcpath))[1][1:].lower()
          if booktype in ['prc', 'mobi', 'pobi', 'azw', 'azw1', 'azw3', 'azw4',
                          'tpz', 'kfx-zip']:
            # Kindle/Mobipocket
            decrypted_ebook = dedrm.KindleMobiDecrypt(srcpath)
          elif booktype == 'pdb':
            # eReader
            decrypted_ebook = dedrm.eReaderDecrypt(srcpath)
          elif booktype == 'pdf':
            # Adobe PDF (hopefully) or LCP PDF
            decrypted_ebook = dedrm.PDFDecrypt(srcpath)
          elif booktype == 'epub':
            # Adobe Adept, PassHash (B&N) or LCP ePub
            decrypted_ebook = dedrm.ePubDecrypt(srcpath)
          else:
            raise ValueError(
                "cannot decrypt %r, unhandle book type %r" %
                (srcpath, booktype)
            )
          assert decrypted_ebook == T.name

  @contextmanager
  def removed(self, srcpath):
    ''' Context manager to produce a deDRMed copy of `srcpath`,
        yielding the temporary file containing the copy.
    '''
    srcext = splitext(basename(srcpath))[1]
    with NamedTemporaryFile(
        prefix=f'.{self.__class__.__name__}-',
        suffix=srcext,
    ) as T:
      self.remove(srcpath, T.name, exists_ok=True)
      yield T.name

  @property
  def kindlekeys(self) -> List[dict]:
    ''' The cached `kindlekeys()`, a list of dicts.
        Obtained via `self.cached_kindlekeys()`.
    '''
    return self.cached_kindlekeys()

  def cached_kindlekeys(self) -> List[dict]:
    ''' Return the cached `kindlekeys()`, a list of dicts.

        If this is empty, a call is made to `self.update_kindlekeys()`
        to load keys from the current environment.
    '''
    kk_tags = self.tags['kindlekeys']
    kks = kk_tags.get('cache')
    if not kks:
      kks = self.update_kindlekeys()
    return kks

  def update_kindlekeys(self, filepaths: Optional[List[str]] = None):
    ''' Update the cached kindle keys from the current environment
        by calling `dedrm.kindlekey.kindlekeys(filepaths)`.
    '''
    if filepaths is None:
      filepaths = []
    new_kks = self.base_kindlekeys(filepaths)
    return self.update_kindlekeys_from_keys(new_kks)

  def update_kindlekeys_from_keys(self, new_kks: Iterable[dict]):
    ''' Update the cached kindle keys with the key `dict`s
        from the iterable `kindlekeys`.
    '''
    kk_tags = self.tags['kindlekeys']
    kks = kk_tags.get('cache') or []
    for keys in new_kks:
      if keys not in kks:
        kks.append(keys)
    print("update_kindlekeys =>", kks)
    kk_tags['cache'] = kks
    return kks

  @classmethod
  def promote(cls, obj) -> "DeDRMWrapper":
    ''' Promote an object to a `DeDRMWrapper`.

        If `obj` is `None` or a `str` return `DeDRMWrapper(dedrm_package_path=obj)`.
    '''
    if isinstance(obj, cls):
      return obj
    if obj is None or isinstance(obj, str):
      return cls(dedrm_package_path=obj)
    raise TypeError("%s.promote: cannot promote %s", cls.__name__, r(obj))

class DeDRMOverride:
  ''' A collection of override methods from the DeDRM/noDRM `DeDRM` class
      which we use to make it work outside the calibre plugin stuff.
  '''

class DeDRM_PrefsOverride:
  ''' Override methods for the DeDRM/noDRM `prefs.DeDRM_Prefs` class.
  '''

def getLibCrypto():
  ''' The OSX `LibCrypto` implementation as an experiment,
        copied from `Other_Tools/DRM_Key_Scripts/Kindle_for_Mac_and_PC/kindlekey.pyw`.
    '''

  class DrmException(Exception):
    ''' Really DeDRM/noDRM needs this in `__init__`, but instead
        there are distinct classes all over the code, a disaster.
        This class is bug-for-bug compatible.
    '''
    pass

  # interface to needed routines in openssl's libcrypto
  def _load_crypto_libcrypto():
    from ctypes import (
        CDLL, byref, POINTER, c_void_p, c_char_p, c_int, c_long, Structure,
        c_ulong, create_string_buffer, addressof, string_at, cast
    )
    from ctypes.util import find_library

    libcrypto = find_library('crypto')
    if libcrypto is None:
      raise DrmException(u"libcrypto not found")
    libcrypto = CDLL(libcrypto)

    # From OpenSSL's crypto aes header
    #
    # AES_ENCRYPT     1
    # AES_DECRYPT     0
    # AES_MAXNR 14 (in bytes)
    # AES_BLOCK_SIZE 16 (in bytes)
    #
    # struct aes_key_st {
    #    unsigned long rd_key[4 *(AES_MAXNR + 1)];
    #    int rounds;
    # };
    # typedef struct aes_key_st AES_KEY;
    #
    # int AES_set_decrypt_key(const unsigned char *userKey, const int bits, AES_KEY *key);
    #
    # note:  the ivec string, and output buffer are both mutable
    # void AES_cbc_encrypt(const unsigned char *in, unsigned char *out,
    #     const unsigned long length, const AES_KEY *key, unsigned char *ivec, const int enc);

    AES_MAXNR = 14
    c_char_pp = POINTER(c_char_p)
    c_int_p = POINTER(c_int)

    class AES_KEY(Structure):
      _fields_ = [
          ('rd_key', c_long * (4 * (AES_MAXNR + 1))), ('rounds', c_int)
      ]

    AES_KEY_p = POINTER(AES_KEY)

    def F(restype, name, argtypes):
      func = getattr(libcrypto, name)
      func.restype = restype
      func.argtypes = argtypes
      return func

    AES_cbc_encrypt = F(
        None, 'AES_cbc_encrypt',
        [c_char_p, c_char_p, c_ulong, AES_KEY_p, c_char_p, c_int]
    )

    AES_set_decrypt_key = F(
        c_int, 'AES_set_decrypt_key', [c_char_p, c_int, AES_KEY_p]
    )

    # From OpenSSL's Crypto evp/p5_crpt2.c
    #
    # int PKCS5_PBKDF2_HMAC_SHA1(const char *pass, int passlen,
    #                        const unsigned char *salt, int saltlen, int iter,
    #                        int keylen, unsigned char *out);

    PKCS5_PBKDF2_HMAC_SHA1 = F(
        c_int, 'PKCS5_PBKDF2_HMAC_SHA1',
        [c_char_p, c_ulong, c_char_p, c_ulong, c_ulong, c_ulong, c_char_p]
    )

    class LibCrypto(object):

      def __init__(self):
        self._blocksize = 0
        self._keyctx = None
        self._iv = 0

      def set_decrypt_key(self, userkey, iv):
        self._blocksize = len(userkey)
        if (self._blocksize != 16) and (self._blocksize !=
                                        24) and (self._blocksize != 32):
          raise DrmException(u"AES improper key used")
          return
        keyctx = self._keyctx = AES_KEY()
        self._iv = iv
        self._userkey = userkey
        rv = AES_set_decrypt_key(userkey, len(userkey) * 8, keyctx)
        if rv < 0:
          raise DrmException(u"Failed to initialize AES key")

      def decrypt(self, data):
        out = create_string_buffer(len(data))
        mutable_iv = create_string_buffer(self._iv, len(self._iv))
        keyctx = self._keyctx
        rv = AES_cbc_encrypt(data, out, len(data), keyctx, mutable_iv, 0)
        if rv == 0:
          raise DrmException(u"AES decryption failed")
        return out.raw

      def keyivgen(self, passwd, salt, iter, keylen):
        saltlen = len(salt)
        passlen = len(passwd)
        out = create_string_buffer(keylen)
        rv = PKCS5_PBKDF2_HMAC_SHA1(
            passwd, passlen, salt, saltlen, iter, keylen, out
        )
        return out.raw

    return LibCrypto

  def _load_crypto():
    LibCrypto = None
    try:
      LibCrypto = _load_crypto_libcrypto()
    except (ImportError, DrmException):
      pass
    return LibCrypto

  LibCrypto = _load_crypto()

  return LibCrypto

if __name__ == '__main__':
  sys.exit(main(sys.argv))



#!/usr/bin/env python3

''' A command line mode for the `cs.ebooks` stuff.
'''

import sys

from cs.cmdutils import BaseCommand

from .apple import AppleBooksCommand
from .calibre import CalibreCommand
from .dedrm import DeDRMCommand
from .kindle import KindleCommand
from .mobi import MobiCommand

class EBooksCommand(BaseCommand):
  ''' Ebooks utility command.
  '''

  cmd_apple = AppleBooksCommand
  cmd_calibre = CalibreCommand
  cmd_dedrm = DeDRMCommand
  cmd_kindle = KindleCommand
  cmd_mobi = MobiCommand

sys.exit(EBooksCommand(sys.argv).run())



#!/usr/bin/env python3
#
# pylint: disable=too-many-lines

''' Support for Calibre libraries.
'''

from code import interact
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timezone
import filecmp
from functools import total_ordering
from getopt import GetoptError
from itertools import chain
import json
import os
from os.path import (
    basename,
    dirname,
    exists as existspath,
    expanduser,
    isabs as isabspath,
    isdir as isdirpath,
    isfile as isfilepath,
    join as joinpath,
    splitext,
)
import re
import shlex
from subprocess import DEVNULL
import sys
from tempfile import TemporaryDirectory
from typing import Optional

from icontract import require
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    Float,
    ForeignKey,
    Integer,
    String,
)
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import declared_attr, relationship
from typeguard import typechecked

from cs.cmdutils import BaseCommand
from cs.context import contextif
from cs.deco import cachedmethod
from cs.fs import FSPathBasedSingleton, HasFSPath, shortpath
from cs.lex import (
    cutprefix,
    get_dotted_identifier,
    lc_,
    FormatableMixin,
    FormatAsError,
)
from cs.logutils import warning, error
from cs.numeric import intif
from cs.obj import SingletonMixin
from cs.pfx import Pfx, pfx_call, pfx_method
from cs.progress import progressbar
from cs.psutils import run
from cs.resources import MultiOpenMixin, RunState, uses_runstate
from cs.sqlalchemy_utils import (
    ORM, BasicTableMixin, HasIdMixin, RelationProxy, proxy_on_demand_field
)
from cs.tagset import TagSet
from cs.threads import locked
from cs.units import transcribe_bytes_geek
from cs.upd import UpdProxy, uses_upd, print  # pylint: disable=redefined-builtin

from .dedrm import DeDRMWrapper, DEDRM_PACKAGE_PATH_ENVVAR

class CalibreTree(FSPathBasedSingleton, MultiOpenMixin):
  ''' Work with a Calibre ebook tree.
  '''

  # used by FSPathBasedSingleton for the default library path
  FSPATH_DEFAULT = '~/CALIBRE'
  FSPATH_ENVVAR = 'CALIBRE_LIBRARY'

  CALIBRE_BINDIR_DEFAULT = '/Applications/calibre.app/Contents/MacOS'

  # pylint: disable=too-many-statements
  @typechecked
  def __init__(self, calibrepath: Optional[str]):
    super().__init__(calibrepath)

    # define the proxy classes
    class CalibreBook(SingletonMixin, RelationProxy(self.db.books, [
        'author_sort',
        'authors',
        'flags',
        'has_cover',
        'isbn',
        'last_modified',
        'lccn',
        'path',
        'pubdate',
        'series_index',
        'sort',
        'tags',
        'timestamp',
        'title',
        'uuid',
    ]), HasFSPath, FormatableMixin):
      ''' A reference to a book in a Calibre library.
      '''

      # pylint: disable=unused-argument
      @classmethod
      def _singleton_key(cls, tree: CalibreTree, dbid: int, db_book=None):
        ''' The singleton key is `(tree,dbid)`.
        '''
        return id(tree), dbid

      @typechecked
      def __init__(self, tree: CalibreTree, dbid: int, db_book=None):
        if 'tree' in self.__dict__:
          if db_book is not None:
            self.refresh_from_db(db_row=db_book)
        else:
          super().__init__(dbid, db_row=db_book)
          self.tree = tree

      def __str__(self):
        return f"{self.dbid}: {self.title}"

      @property
      def fspath(self):
        ''' The filesystem path to where the book formats are stored.
        '''
        return self.tree.pathto(self.path)

      def format_tagset(self):
        ''' Compute a `TagSet` representing this book's metadata.
        '''
        formats = self.formats
        tags = TagSet(
            dbid=self.dbid,
            asin=self.asin,
            title=self.title,
            authors=self.author_names,
            formats=formats,
            format_names=sorted(formats.keys()),
            tags=self.tags,
        )
        tags.update(
            {
                'series.id': self.series_id,
                'series.name': self.series_name,
                'series.index': intif(self.series_index),
            }
        )
        return tags

      def format_kwargs(self):
        return self.format_tagset()

      def get_arg_name(self, field_name):
        ''' Override for `FormattableMixin.get_arg_name`:
            return the leading dotted identifier,
            which represents a tag or tag prefix.
        '''
        return get_dotted_identifier(field_name)

      @property
      def dbid(self):
        ''' An alias for the `.id` attribute.
        '''
        return self.id

      @property
      def asin(self):
        ''' The Amazon ASIN, or `None`, from `self.identifiers['mobi-asin']`.
        '''
        return self.identifiers.get('mobi-asin', None)

      @classmethod
      def refresh_from_db_row(cls, db_row, fields, *, session):
        ''' Refresh the cached values from the database.
        '''
        super().refresh_from_db_row(db_row, fields, session=session)
        for field_name in 'authors', 'formats', 'idenitifiers', 'tags':
          try:
            del fields[field_name]
          except KeyError:
            pass

      @property
      @proxy_on_demand_field
      # pylint: disable=property-with-parameters
      def authors(self, db_row, *, session):
        ''' The book Authors.
        '''
        return db_row.authors

      @property
      @proxy_on_demand_field
      # pylint: disable=property-with-parameters
      def series_id(self, db_row, *, session):
        ''' The book Series or `None`.
        '''
        if not db_row.series:
          return None
        sid, = db_row.series
        return sid

      @property
      def series_name(self):
        ''' The series name or `None`.
        '''
        sid = self.series_id
        if sid is None:
          return None
        orm = self.orm
        with orm.session() as session:
          srow = orm.series.by_id(sid, session=session)
          return srow.name

      @property
      def author_names(self):
        ''' A list of the author names.
        '''
        return [author.name for author in self.authors]

      @property
      @proxy_on_demand_field
      # pylint: disable=property-with-parameters
      def formats(self, db_row, *, session):
        ''' A mapping of Calibre format keys to format paths
            computed on demand.
        '''
        return {
            fmt.format: f'{fmt.name}.{fmt.format.lower()}'
            for fmt in db_row.formats
        }

      def formatpath(self, fmtk):
        ''' Return the filesystem path of the format file for `fmtk`
            or `None` if the format is not present.
        '''
        try:
          fmtsubpath = self.formats[fmtk]
        except KeyError:
          return None
        return self.pathto(fmtsubpath)

      @pfx_method
      @typechecked
      def add_format(
          self,
          bookpath: str,
          *,
          force: bool = False,
          doit=True,
          quiet=False,
          **subp_options,
      ):
        ''' Add a book file to the existing book formats
            via the `calibredb add_format` command.
            Return `True` if the `doit` is false or the command succeeds,
            `False` otherwise.

            Parameters:
            * `bookpath`: filesystem path to the source MOBI file
            * `doit`: default `True`; do not run the command if false
            * `force`: replace an existing format if already present, default `False`
            * `quiet`: default `False`; only print warning if true
        '''
        cp = self.tree.calibredb(
            'add_format',
            *(() if force else ('--dont-replace',)),
            str(self.id),
            bookpath,
            doit=doit,
            quiet=quiet,
            stdin=DEVNULL,
            **subp_options,
        )
        if cp is None:
          return True
        if cp.returncode != 0:
          warning("command fails, return code %d", cp.returncode)
          return False
        self.refresh_from_db()
        return True

      @property
      @proxy_on_demand_field
      # pylint: disable=property-with-parameters
      def identifiers(self, db_row, *, session):
        ''' A mapping of Calibre identifier keys to identifier values
            computed on demand.
        '''
        return {
            identifier.type: identifier.val
            for identifier in db_row.identifiers
        }

      @property
      @proxy_on_demand_field
      # pylint: disable=property-with-parameters
      def tags(self, db_row, *, session):
        ''' A list of Calibre tags computed on demand.
        '''
        return [tag.name for tag in db_row.tags]

      # TODO: should really edit the db directly
      @tags.setter
      def tags(self, new_tags):
        ''' Update the tags.
        '''
        self.tree.calibredb(
            'set_metadata',
            '--field',
            f'tags:{",".join(new_tags)}',
            str(self.dbid),
        )
        try:
          del self._RelProxy__fields['tags']
        except KeyError:
          pass

      def convert(
          self,
          srcfmtk,
          dstfmtk,
          *conv_opts,
          doit=True,
          force=False,
          quiet=False,
      ):
        ''' Convert the existing format `srcfmtk` into `dstfmtk`.
        '''
        calibre = self.tree
        srcpath = self.formatpath(srcfmtk)
        srcbase = basename(srcpath)
        dstbase = splitext(srcbase)[0] + '.' + dstfmtk.lower()
        if srcbase == dstbase:
          raise ValueError(
              "source format basename %r == destination format basename %r, skipping"
              % (srcbase, dstbase)
          )
        with TemporaryDirectory(prefix='.convert-') as dirpath:
          dstpath = joinpath(dirpath, dstbase)
          calibre.ebook_convert(
              srcpath, dstpath, *conv_opts, check=True, doit=doit, quiet=quiet
          )
          self.add_format(dstpath, force=force, doit=doit, quiet=quiet)

      @property
      def mobipath(self):
        ''' The filesystem path of a Mobi format book file, or `None`.
        '''
        for fmtk in 'MOBI', 'AZW3', 'AZW':
          fmtpath = self.formatpath(fmtk)
          if fmtpath is not None:
            return fmtpath
        return None

      def make_cbz(self, replace_format=False):
        ''' Create a CBZ format from the AZW3 Mobi format.
        '''
        from .mobi import Mobi  # pylint: disable=import-outside-toplevel
        formats = self.formats
        if 'CBZ' in formats and not replace_format:
          warning("format CBZ already present, not adding")
        else:
          mobipath = self.mobipath
          if mobipath:
            base, _ = splitext(basename(mobipath))
            MB = Mobi(mobipath)
            with TemporaryDirectory() as tmpdirpath:
              cbzpath = joinpath(tmpdirpath, base + '.cbz')
              pfx_call(MB.make_cbz, cbzpath)
              self.add_format(cbzpath, force=replace_format)
          else:
            raise ValueError(
                "no AZW3, AZW or MOBI format from which to construct a CBZ"
            )

      @uses_runstate
      def pull(
          self,
          obook,
          *,
          doit=True,
          formats=None,
          force=False,
          quiet=False,
          verbose=False,
          runstate: RunState,
      ):
        ''' Pull formats from another `CalibreBook`.

            Parameters:
            * `obook`: the other book
            * `doit`: optional flag, default `True`;
              import formats if true, report actions otherwise
            * `formats`: optional list of Calibre format keys to pull if present
            * `force`: optional flag, default `False`;
              if true import formats even if already present
            * `quiet`: optional flag, default `False`;
              if true only print warnings
            * `verbose`: optional flag, default `False`;
              if true print all actions and inactions
        '''
        if formats is None:
          formats = sorted(obook.formats.keys())
        with Pfx("%s <= %s", self, obook):
          for fmtk in formats:
            if runstate.cancelled:
              break
            ofmtpath = obook.formatpath(fmtk)
            if ofmtpath is None:
              continue
            with Pfx(fmtk):
              self.pull_format(
                  ofmtpath,
                  fmtk=fmtk,
                  doit=doit,
                  force=force,
                  quiet=quiet,
                  verbose=verbose
              )

      def pull_format(
          self,
          ofmtpath,
          *,
          fmtk=None,
          doit=True,
          force=False,
          quiet=False,
          verbose=False,
      ):
        ''' Pull a format file, typically from another `CalibreBook`.

            Parameters:
            * `ofmtpath`: the filesystem path of the format to pull
            * `fmtk`: optional format key,
              default derived from the `ofmtpath` filename extension
            * `doit`: optional flag, default `True`;
              import formats if true, report actions otherwise
            * `force`: optional flag, default `False`;
              if true import formats even if already present
            * `quiet`: optional flag, default `False`;
              if true only print warnings
            * `verbose`: optional flag, default `False`;
              if true print all actions and inactions
        '''
        if fmtk is None:
          _, ext = splitext(basename(ofmtpath))
          if ext:
            assert ext.startswith('.')
            fmtk = ext[1:].upper()
        if fmtk is None:
          warning(
              "cannot infer format key from %r, not doing a precheck", ofmtpath
          )
        else:
          fmtpath = self.formatpath(fmtk)
          if fmtpath is None and fmtk.startswith('AZW'):
            fmtpath = (
                self.formatpath('AZW4') or self.formatpath('AZW3')
                or self.formatpath('AZW') or self.formatpath('MOBI')
            )
          if fmtpath is not None and not force:
            if filecmp.cmp(fmtpath, ofmtpath):
              # pylint: disable=expression-not-assigned
              verbose and print(
                  self, fmtk, "identical to", shortpath(ofmtpath)
              )
            else:
              # pylint: disable=expression-not-assigned
              verbose and warning(
                  "already present with different content\n"
                  "  present: %s\n"
                  "  other:   %s",
                  shortpath(fmtpath),
                  shortpath(ofmtpath),
              )
            return
        # pylint: disable=expression-not-assigned
        quiet or print(
            self, self.formats, '+', fmtk, '<=', shortpath(ofmtpath)
        )
        self.add_format(ofmtpath, doit=doit, force=force, quiet=quiet)

    self.CalibreBook = CalibreBook

  def __str__(self):
    return "%s:%s" % (type(self).__name__, self.shortpath)

  @contextmanager
  def startup_shutdown(self):
    ''' Stub startup/shutdown.
    '''
    yield

  @property
  @locked
  @cachedmethod
  def db(self):
    ''' The associated `CalibreMetadataDB` ORM,
        instantiated on demand.
    '''
    return CalibreMetadataDB(self)

  @contextmanager
  def db_session(self):
    ''' Context manager for a database session.
        This is not fired by `startup_shutdown` because any callout
        to the Calibre command executables also want exclusive
        access to the database.
    '''
    with self.db.session() as session:
      yield session

  def dbshell(self):
    ''' Interactive db shell.
    '''
    return self.db.shell()

  @uses_upd
  def preload(self, upd):
    ''' Scan all the books, preload their data.
    '''
    with upd.run_task(f'preload {self}'):
      db = self.db
      with db.session() as session:
        for db_book in self.db.books.lookup(session=session):
          self.book_by_dbid(db_book.id, db_book=db_book)

  @typechecked
  def __getitem__(self, dbid: int):
    return self.book_by_dbid(dbid)

  def __contains__(self, dbid: int):
    db = self.db
    try:
      with db.session() as session:
        db.books.by_id(dbid, session=session)
    except IndexError:
      return False
    return True

  @typechecked
  @require(lambda dbid: dbid > 0)
  def book_by_dbid(self, dbid: int, *, db_book=None):
    ''' Return a cached `CalibreBook` for `dbid`.
    '''
    return self.CalibreBook(self, dbid, db_book=db_book)

  def __iter__(self):
    ''' Generator yielding `CalibreBook`s.
    '''
    db = self.db
    seen_dbids = set()
    with db.session() as session:
      for author in sorted(db.authors.lookup(session=session)):
        with Pfx("%d:%s", author.id, author.name):
          for book in sorted(author.books):
            if book.id in seen_dbids:
              continue
            yield self.book_by_dbid(book.id, db_book=book)
            seen_dbids.add(book.id)

  def identifier_names(self):
    ''' Return an iterable of the identifiers in use in the library.
    '''
    return set(chain(*(cbook.identifiers.keys() for cbook in self)))

  def by_identifier(self, type_, value):
    ''' Generator yielding `CalibreBook`
        matching the provided `(type,val)` identifier.
    '''
    db = self.db
    with db.session() as session:
      for identifier in db.identifiers.lookup(session=session, type=type_,
                                              val=value):
        yield self[identifier.book_id]

  def by_asin(self, asin):
    ''' Return an iterable of `CalibreBook`s with the supplied ASIN.
    '''
    return self.by_identifier('mobi-asin', asin.upper())

  def _run(self, calcmd, *calargv, doit=True, quiet=False, **subp_options):
    ''' Run a Calibre utility command.
        Return the `CompletedProcess` result.

        Parameters:
        * `calcmd`: the Calibre command to invoke;
          if the command name is not an absolute path
          it is expected to come from `self.CALIBRE_BINDIR_DEFAULT`
        * `calargv`: the arguments for the command
        * `doit`: default `True`; do not run the command of false
        * `quiet`: default `False`; if true, do not print the command or its output
        * `subp_options`: optional mapping of keyword arguments
          to pass to `subprocess.run`
    '''
    subp_options.setdefault('capture_output', True)
    subp_options.setdefault('check', False)
    subp_options.setdefault('text', True)
    if not isabspath(calcmd):
      calcmd = joinpath(self.CALIBRE_BINDIR_DEFAULT, calcmd)
    calargv = [calcmd, *calargv]
    cp = run(calargv, doit=doit, quiet=quiet, **subp_options)
    if cp is not None and cp.stdout and not quiet:
      print(" ", cp.stdout.rstrip().replace("\n", "\n  "))
    return cp

  def calibredb(self, dbcmd, *argv, doit=True, quiet=False, **subp_options):
    ''' Run `dbcmd` via the `calibredb` command.
        Return a `CompletedProcess` or `None` if `doit` is false.
    '''
    subp_argv = [
        'calibredb',
        dbcmd,
        '--library-path=' + self.fspath,
        *argv,
    ]
    return self._run(*subp_argv, doit=doit, quiet=quiet, **subp_options)

  def ebook_convert(
      self,
      srcpath,
      dstpath,
      *conv_opts,
      doit=True,
      quiet=False,
      **subp_options
  ):
    ''' Run `dbcmd` via the `calibredb` command.
        Return a `CompletedProcess` or `None` if `doit` is false.
    '''
    if not isfilepath(srcpath):
      raise ValueError("source path is not a file: %r" % (srcpath,))
    if existspath(dstpath):
      raise ValueError("destination path already exists: %r" % (dstpath,))
    subp_argv = [
        'ebook-convert',
        srcpath,
        dstpath,
        *conv_opts,
    ]
    return self._run(*subp_argv, doit=doit, quiet=quiet, **subp_options)

  @pfx_method
  def add(
      self,
      bookpath,
      *,
      dedrm=None,
      doit=True,
      quiet=False,
      **subp_options,
  ):
    ''' Add a book file via the `calibredb add` command.
        Return the database id or `None` if `doit` is false or the command fails.

        Parameters:
        * `bookpath`: the filesystem path to the book
        * `dedrm`: optional `DeDRMWrapper` instance
    '''
    if dedrm is not None:
      # try to remove DRM from the book file
      # and add the cleared temporary copy
      if doit:
        with dedrm.removed(bookpath) as clearpath:
          return self.add(clearpath, doit=doit, quiet=quiet, **subp_options)
    cp = self.calibredb(
        'add',
        '--duplicates',
        bookpath,
        doit=doit,
        quiet=quiet,
        stdin=DEVNULL,
        capture_output=True,
        text=True,
        **subp_options,
    )
    if cp is None:
      return None
    if cp.returncode != 0:
      return None
    # Extract the database id from the "calibredb add" output.
    dbids = []
    output_lines = cp.stdout.split('\n')
    for line in output_lines:
      line_sfx = cutprefix(line, 'Added book ids:')
      if line_sfx is not line:
        dbids.extend(
            map(lambda dbid_s: int(dbid_s.strip()), line_sfx.split(','))
        )
    if not dbids:
      warning(
          "calibredb add %r failed\n  Output:\n    %s", bookpath,
          "    \n".join(output_lines)
      )
      return None
    dbid, = dbids  # pylint: disable=unbalanced-tuple-unpacking
    return dbid

# pylint: disable=too-many-instance-attributes
class CalibreMetadataDB(ORM):
  ''' An ORM to access the Calibre `metadata.db` SQLite database.
  '''

  DB_FILENAME = 'metadata.db'

  def __init__(self, tree):
    if isinstance(tree, str):
      tree = CalibreTree(tree)
    self.tree = tree
    self.db_url = 'sqlite:///' + self.db_path
    super().__init__(self.db_url)

  @property
  def orm(self):
    ''' No distinct ORM class for `CalibreMetadataDB`.
    '''
    return self

  @property
  def db_path(self):
    ''' The filesystem path to the database.
    '''
    return self.tree.pathto(self.DB_FILENAME)

  def shell(self):
    ''' Interactive db shell.
    '''
    print("sqlite3", self.db_path)
    run(['sqlite3', self.db_path], check=True)
    return 0

  # lifted from SQLTags
  @contextmanager
  def session(self, *, new=False):
    ''' Context manager to obtain a db session if required
        (or if `new` is true).
    '''
    orm_state = self.orm.sqla_state
    get_session = orm_state.new_session if new else orm_state.auto_session
    with get_session() as session2:
      yield session2

  # pylint: disable=too-many-statements
  def declare_schema(self):
    r''' Define the database schema / ORM mapping.

        Database schema queried thus:

            sqlite3 ~/CALIBRE/metadata.db .schema
    '''
    Base = self.Base

    class _CalibreTable(BasicTableMixin, HasIdMixin):
      ''' Base class for Calibre tables.
      '''

    def _linktable(
        left_name, right_name, tablename=None, **additional_columns
    ):
      ''' Prepare and return a Calibre link table base class.

          Parameters:
          * `left_name`: the left hand entity, lowercase, singular,
            example `'book'`
          * `right_name`: the right hand entity, lowercase, singular,
            example `'author'`
          * `additional_columns`: other keyword parameters
            define further `Column`s and relationships
      '''
      if tablename is None:
        tablename = f'{left_name}s_{right_name}s_link'

      class linktable(_CalibreTable):
        ''' Prepare a `_CalibreTable` subclass representing a Calibre link table.
        '''
        __tablename__ = tablename

      def vsetattr(o, a, v):
        ##X("setattr(%r,%r,%r)", o, a, v)
        setattr(o, a, v)

      vsetattr(
          linktable, f'{left_name}_id',
          declared_attr(
              lambda self: Column(
                  left_name,
                  ForeignKey(f'{left_name}s.id'),
                  primary_key=True,
              )
          )
      )
      vsetattr(
          linktable, left_name,
          declared_attr(
              lambda self: relationship(
                  f'{left_name.title()}s',
                  back_populates=f'{right_name}_links',
              )
          )
      )
      vsetattr(
          linktable, f'{right_name}_id',
          declared_attr(
              lambda self: Column(
                  right_name,
                  ForeignKey(f'{right_name}s.id'),
                  primary_key=True,
              )
          )
      )
      vsetattr(
          linktable, right_name,
          declared_attr(
              lambda self: relationship(
                  f'{right_name.title()}s',
                  back_populates=f'{left_name}_links',
              )
          )
      )
      for colname, colspec in additional_columns.items():
        vsetattr(
            linktable,
            colname,
            declared_attr(lambda self, colspec=colspec: colspec),
        )
      return linktable

    @total_ordering
    class Authors(Base, _CalibreTable):
      ''' An author.
      '''
      __tablename__ = 'authors'
      name = Column(String, nullable=False, unique=True)
      sort = Column(String)
      link = Column(String, nullable=False, default="")

      def __hash__(self):
        return self.id

      def __eq__(self, other):
        return self.id == other.id

      def __lt__(self, other):
        return self.sort.lower() < other.sort.lower()

    @total_ordering
    class Books(Base, _CalibreTable):
      ''' A book.
      '''
      __tablename__ = 'books'
      title = Column(String, nullable=False, unique=True, default='unknown')
      sort = Column(String)
      timestamp = Column(DateTime)
      pubdate = Column(DateTime)
      series_index = Column(Float, nullable=False, default=1.0)
      author_sort = Column(String)
      isbn = Column(String, default="")
      lccn = Column(String, default="")
      path = Column(String, nullable=False, default="")
      flags = Column(Integer, nullable=False, default=1)
      uuid = Column(String)
      has_cover = Column(Boolean, default=False)
      last_modified = Column(
          DateTime,
          nullable=False,
          default=datetime(2000, 1, 1, tzinfo=timezone.utc)
      )

      def __hash__(self):
        return self.id

      def __eq__(self, other):
        return self.id == other.id

      def __lt__(self, other):
        return self.author_sort.lower() < other.author_sort.lower()

      def identifiers_as_dict(self):
        ''' Return a `dict` mapping identifier types to values.
        '''
        return {
            identifier.type: identifier.val
            for identifier in self.identifiers
        }

    class Data(Base, _CalibreTable):
      ''' Data files associated with a book.
      '''
      __tablename__ = 'data'
      book_id = Column(
          "book", ForeignKey('books.id'), nullable=False, primary_key=True
      )
      format = Column(String, nullable=False, primary_key=True)
      uncompressed_size = Column(Integer, nullable=False)
      name = Column(String, nullable=False)

    class Identifiers(Base, _CalibreTable):
      ''' Identifiers associated with a book such as `"isbn"` or `"mobi-asin"`.
      '''
      __tablename__ = 'identifiers'
      book_id = Column("book", ForeignKey('books.id'), nullable=False)
      type = Column(String, nullable=False, default="isbn")
      val = Column(String, nullable=None)

    class Languages(Base, _CalibreTable):
      ''' Lamguage codes.
      '''
      __tablename__ = 'languages'
      lang_code = Column(String, nullable=False, unique=True)

    class Preferences(Base, _CalibreTable):
      ''' Calibre preferences.
      '''
      __tablename__ = 'preferences'
      key = Column(String, nullable=False, unique=True)
      value = Column("val", String, nullable=False)

    class Series(Base, _CalibreTable):
      ''' Calibre preferences.
      '''
      __tablename__ = 'series'
      name = Column(String, nullable=False, unique=True)
      sort = Column(String, nullable=False)

    @total_ordering
    class Tags(Base, _CalibreTable):
      ''' A tag.
      '''
      __tablename__ = 'tags'
      name = Column(String, nullable=False, unique=True)

      def __hash__(self):
        return self.id

      def __eq__(self, other):
        return self.id == other.id

      def __lt__(self, other):
        return self.name.lower() < other.name.lower()

    class BooksAuthorsLink(Base, _linktable('book', 'author')):
      ''' Link table between `Books` and `Authors`.
      '''

    class BookSeriesLink(Base, _CalibreTable):
      ''' Link table between `Books` and `Series`.
      '''
      __tablename__ = 'books_series_link'
      book = Column('book', ForeignKey('books.id'))
      series = Column('series', ForeignKey('series.id'))

    class BooksTagsLink(Base, _linktable('book', 'tag')):
      ''' Link table between `Books` and `Tags`.
      '''

    ##class BooksLanguagesLink(Base, _linktable('book', 'lang_code')):
    ##  item_order = Column(Integer, nullable=False, default=1)

    Authors.book_links = relationship(BooksAuthorsLink)
    Authors.books = association_proxy('book_links', 'book')

    Books.author_links = relationship(BooksAuthorsLink)
    Books.authors = association_proxy('author_links', 'author')
    Books.formats = relationship(Data, backref="book")
    Books.identifiers = relationship(Identifiers)
    Books.series_links = relationship(BookSeriesLink)
    Books.series = association_proxy('series_links', 'series')
    Books.tag_links = relationship(BooksTagsLink)
    Books.tags = association_proxy('tag_links', 'tag')

    ##Books.language_links = relationship(BooksLanguagesLink)
    ##Books.languages = association_proxy('languages_links', 'languages')

    Identifiers.book = relationship(Books, back_populates="identifiers")

    Tags.book_links = relationship(BooksTagsLink)
    Tags.books = association_proxy('book_links', 'book')

    # references to table definitions
    self.authors = Authors
    Authors.orm = self
    self.books = Books
    Books.orm = self
    self.identifiers = Identifiers
    Identifiers.orm = self
    self.languages = Languages
    Languages.orm = self
    self.preferences = Preferences
    Preferences.orm = self
    self.series = Series
    Series.orm = self
    self.tags = Tags
    Tags.orm = self

class CalibreCommand(BaseCommand):
  ''' Command line tool to interact with a Calibre filesystem tree.
  '''

  GETOPT_SPEC = 'C:K:O:'

  USAGE_FORMAT = '''Usage: {cmd} [-C calibre_library] [-K kindle-library-path] subcommand [...]
  -C calibre_library
    Specify calibre library location.
  -K kindle_library
    Specify kindle library location.
  -O other_calibre_library
    Specify alternate calibre library location, the default library
    for pull etc. The default comes from ${OTHER_LIBRARY_PATH_ENVVAR}.'''

  # envar $CALIBRE_LIBRARY_OTHER as push/pull etc "other library"
  OTHER_LIBRARY_PATH_ENVVAR = CalibreTree.FSPATH_ENVVAR + '_OTHER'

  SUBCOMMAND_ARGV_DEFAULT = 'info'

  DEFAULT_LINK_IDENTIFIER = 'mobi-asin'

  DEFAULT_LINKTO_DIRPATH = '~/media/books'
  DEFAULT_LINKTO_DIRPATH_ENVVAR = 'MEDIA_BOOKSPATH'
  DEFAULT_LINKTO_FORMATS = ['CBZ', 'EPUB']
  DEFAULT_LINKTO_SELECTORS = ['CBZ', 'EPUB']
  DEFAULT_LINKTO_SERIES_FORMAT = '{series.name:lc}--{series.index}--{title:lc}'
  DEFAULT_LINKTO_NOSERIES_FORMAT = '{title:lc}'

  USAGE_KEYWORDS = {
      'DEFAULT_LINK_IDENTIFIER': DEFAULT_LINK_IDENTIFIER,
      'OTHER_LIBRARY_PATH_ENVVAR': OTHER_LIBRARY_PATH_ENVVAR,
      'DEFAULT_LINKTO_DIRPATH': DEFAULT_LINKTO_DIRPATH,
      'DEFAULT_LINKTO_DIRPATH_ENVVAR': DEFAULT_LINKTO_DIRPATH_ENVVAR,
      'DEFAULT_LINKTO_FORMATS': DEFAULT_LINKTO_FORMATS,
      'DEFAULT_LINKTO_SELECTORS': DEFAULT_LINKTO_SELECTORS,
      'DEFAULT_LINKTO_SERIES_FORMAT': DEFAULT_LINKTO_SERIES_FORMAT,
      'DEFAULT_LINKTO_NOSERIES_FORMAT': DEFAULT_LINKTO_NOSERIES_FORMAT,
  }

  # mapping of target format key to source format and extra options
  CONVERT_MAP = {
      'EPUB': (['MOBI', 'AZW', 'AZW3'], ()),
  }

  @dataclass
  class Options(BaseCommand.Options):
    ''' Special class for `self.options` with various properties.
    '''

    calibre_path: str = field(
        default_factory=lambda: CalibreTree.
        _resolve_fspath(None, CalibreTree.FSPATH_ENVVAR)
    )
    calibre_path_other: Optional[str] = field(
        default_factory=lambda: (
            CalibreTree.
            _resolve_fspath(None, CalibreCommand.OTHER_LIBRARY_PATH_ENVVAR)
            if CalibreCommand.OTHER_LIBRARY_PATH_ENVVAR in os.environ else None
        )
    )
    dedrm_package_path: Optional[str] = field(
        default_factory=lambda: os.environ.get(DEDRM_PACKAGE_PATH_ENVVAR)
    )

    def _default_kindle_path():
      from .kindle import KindleTree  # pylint: disable=import-outside-toplevel
      return (
          KindleTree._resolve_fspath(None)
          if KindleTree.FSPATH_ENVVAR in os.environ else None
      )

    kindle_path: Optional[str] = field(default_factory=_default_kindle_path)
    linkto_dirpath: str = field(
        default_factory=lambda: os.environ.get(
            CalibreCommand.DEFAULT_LINKTO_DIRPATH_ENVVAR
        ) or expanduser(CalibreCommand.DEFAULT_LINKTO_DIRPATH)
    )

    @property
    def calibre(self):
      ''' The `CalibreTree` from `self.calibre_path`.
      '''
      return CalibreTree(self.calibre_path)

    @property
    def calibre_other(self):
      ''' The alternate `CalibreTree` from `self.calibre_path_other`.
      '''
      if self.calibre_path_other is None:
        raise AttributeError(".calibre_other: no .calibre_path_other")
      return CalibreTree(self.calibre_path_other)

    @property
    def kindle(self):
      ''' The `KindleTree` from `self.kindle_path`.
      '''
      if self.kindle_path is None:
        raise AttributeError(".kindle: no .kindle_path")
      from .kindle import KindleTree  # pylint: disable=import-outside-toplevel
      return KindleTree(self.kindle_path)

  def apply_opt(self, opt, val):
    ''' Apply a command line option.
    '''
    options = self.options
    if opt == '-C':
      options.calibre_path = val
    elif opt == '-K':
      options.kindle_path = val
    elif opt == '-O':
      options.calibre_path_other = val
    else:
      super().apply_opt(opt, val)

  @contextmanager
  def run_context(self):
    ''' Prepare the `SQLTags` around each command invocation.
    '''
    with super().run_context():
      calibre = self.options.calibre
      with calibre:
        with calibre.db_session():
          yield

  @staticmethod
  def books_from_spec(calibre, book_spec):
    ''' Generator yielding `CalibreBook` instances from `book_spec`.
    '''
    # raw dbid
    try:
      dbid = int(book_spec)
    except ValueError:
      # FORMAT
      if book_spec.isupper():
        match_fn = lambda book: book_spec in book.formats
      # /regexp
      elif book_spec.startswith('/'):
        re_s = book_spec[1:]
        if not re_s:
          raise ValueError("empty regexp")  # pylint: disable=raise-missing-from
        regexp = re.compile(re_s, re.I)
        match_fn = lambda book: (
            regexp.search(book.title) or any(
                map(regexp.search, book.author_names)
            ) or regexp.search(book.series_name or "") or
            any(map(regexp.search, book.tags))
        )
      else:
        # [identifier=]id-value,...
        try:
          identifiers_s, values_s = book_spec.split('=', 1)
        except ValueError:
          # id-value,...
          identifiers = None
          values_s = book_spec
        else:
          identifiers = identifiers_s.split(',')
        values = list(map(str.lower, values_s.split(',')))
        match_fn = lambda book: any(
            (
                (identifiers is None or idk in identifiers) and idv.lower() in
                values
            ) for idk, idv in book.identifiers.items()
        )
      for book in calibre:
        if match_fn(book):
          yield book
    else:
      yield calibre[dbid]

  @staticmethod
  def cbook_default_sortkey(cbook):
    ''' The default presentation order for things like "ls":
        series-name-index,title,authors,dbid.
    '''
    return (
        (cbook.series_name.lower(),
         cbook.series_index) if cbook.series_name else ("", 0),
        cbook.title.lower(),
        tuple(map(str.lower, cbook.author_names)),
        cbook.dbid,
    )

  def popbooks(self, argv, once=False, sortkey=None, reverse=False):
    ''' Consume `argv` as book specifications and return a list of matching books.

        If `once` is true (default `False`) consume only the first argument.
        If `sortkey` is not `None`, sort the returned books by this function.
    '''
    options = self.options
    calibre = options.calibre
    cbooks = []
    while argv:
      book_spec = self.poparg(argv, "book_spec")
      cbooks.extend(self.books_from_spec(calibre, book_spec))
      if once:
        break
    if sortkey is not None and sortkey is not False:
      if sortkey is True:
        sortkey = self.cbook_default_sortkey
      cbooks = sorted(cbooks, key=sortkey, reverse=reverse)
    return cbooks

  def cmd_add(self, argv):
    ''' Usage: {cmd} [-nqv] bookpaths...
          Add the specified ebook bookpaths to the library.
          -n    No action: recite planned actions.
          -q    Quiet: only emit warnings.
          -v    Verbose: report all actions and decisions.
    '''
    options = self.options
    dedrm = (
        DeDRMWrapper(options.dedrm_package_path)
        if options.dedrm_package_path else None
    )
    calibre = options.calibre
    self.popopts(argv, options, n='doit', q='quiet', v='verbose')
    if not argv:
      raise GetoptError("missing bookpaths")
    for bookpath in argv:
      with Pfx(bookpath):
        calibre.add(
            bookpath,
            dedrm=dedrm,
            doit=options.doit,
            quiet=options.quiet,
        )

  # pylint: disable=too-many-branches,too-many-locals
  def cmd_convert(self, argv):
    ''' Usage: {cmd} [-fnqv] formatkey dbids...
          Convert books to the format `formatkey`.
          -f    Force: convert even if the format is already present.
          -n    No action: recite planned actions.
          -q    Quiet: only emit warnings.
          -v    Verbose: report all actions and decisions.
    '''
    options = self.options
    self.popopts(argv, options, f='force', n='doit', q='quiet', v='verbose')
    dstfmtk = self.poparg(argv).upper()
    srcfmtks, conv_opts = self.CONVERT_MAP.get(dstfmtk, ([], ()))
    if not srcfmtks:
      raise GetoptError(
          "no source formats can produce formatkey %r" % (dstfmtk,)
      )
    if not argv:
      raise GetoptError("missing dbids")
    try:
      cbooks = self.popbooks(argv)
    except ValueError as e:
      raise GetoptError("invalid book specifiers: %s") from e
    xit = 0
    doit = options.doit
    force = options.force
    quiet = options.quiet
    verbose = options.verbose
    runstate = options.runstate
    for cbook in cbooks:
      if runstate.cancelled:
        break
      with Pfx(cbook):
        if dstfmtk in cbook.formats:
          if force:
            # pylint: disable=expression-not-assigned
            verbose and warning("replacing format %r")
          else:
            # pylint: disable=expression-not-assigned
            verbose and print(f"{cbook}: format {dstfmtk!r} already present")
            continue
        for srcfmtk in srcfmtks:
          if srcfmtk in cbook.formats:
            break
        else:
          srcfmtk = None
        if srcfmtk is None:
          warning(
              "no suitable source formats (%r); I looked for %r",
              sorted(cbook.formats.keys()), srcfmtks
          )
          xit = 1
          continue

      cbook.convert(srcfmtk, dstfmtk, *conv_opts, doit=doit, quiet=quiet)

    if runstate.cancelled:
      xit = 1
    return xit

  def cmd_dbshell(self, argv):
    ''' Usage: {cmd}
          Start an interactive database prompt.
    '''
    if argv:
      raise GetoptError("extra arguments: %r" % (argv,))
    return self.options.calibre.dbshell()

  def cmd_info(self, argv):
    ''' Usage: {cmd}
          Report basic information.
    '''
    if argv:
      raise GetoptError("extra arguments: %r" % (argv,))
    print("calibre", self.options.calibre.shortpath)
    if self.options.calibre_path_other:
      print("calibre_other", shortpath(self.options.calibre_path_other))
    if self.options.kindle_path:
      print("kindle", shortpath(self.options.kindle_path))

  def cmd_linkto(self, argv):
    ''' Usage: {cmd} [-1fnqv] [-d linkto-dir] [-F fmt,...] [-o link-format] [dbids...]
          Export books to linkto-dir by hard linking.
          -1              Link only the first format found.
          -d linkto-dir   Specify the target directory, default from ${DEFAULT_LINKTO_DIRPATH_ENVVAR}
                          or {DEFAULT_LINKTO_DIRPATH}.
          -F fmt,...      Source formats, default: {DEFAULT_LINKTO_FORMATS}
          -f              Force. Replace existing links.
          -n              No action. Report planned actions.
          -o link-format  Link name format.
          -q              Quiet.
          -v              Verbose.
    '''
    ##Default with series: {DEFAULT_LINKTO_SERIES_FORMAT}
    ##Default without series: {DEFAULT_LINKTO_NOSERIES_FORMAT}
    options = self.options
    options.formats = ['CBZ', 'EPUB']
    options.first_format = False
    options.link_format = None
    self.popopts(
        argv,
        options,
        _1='first_format',
        d_='linkto_dirpath',
        F_='formats',
        f='force',
        n='-doit',
        o_='link_format',
        q='quiet',
        v='verbose',
    )
    doit = options.doit
    first_format = options.first_format
    force = options.force
    calibre = options.calibre
    formats = options.formats
    if isinstance(formats, str):
      # pylint: disable=no-member
      formats = [fmt.strip().upper() for fmt in formats.split(',')]
    link_format = options.link_format
    linkto_dirpath = options.linkto_dirpath
    quiet = options.quiet
    runstate = options.runstate
    verbose = options.verbose
    upd = options.upd
    quiet or print(
        "linkto", calibre.shortpath, "=>", shortpath(options.linkto_dirpath)
    )
    cbooks = sorted(
        set(self.popbooks(argv or list(self.DEFAULT_LINKTO_SELECTORS))),
        key=lambda cbook: cbook.title.lower()
    )
    for cbook in progressbar(cbooks, "linkto"):
      with upd.run_task('linkto: ') as proxy:
        if runstate.cancelled:
          break
        proxy.text = str(cbook)
        with Pfx(cbook):
          fmttags = cbook.format_tagset()
          series_name = fmttags.get('series.name')
          if series_name:
            series_name = series_name.strip()
          name_format = link_format or (
              '{series.name:lc}--{series.index}--{title:lc}'
              if series_name else '{title:lc}'
          )
          name = (
              cbook.format_as(name_format).replace('_', '-').replace('/', ':')
          )
          name_subdir = (
              lc_(series_name).replace('/', ':')
              if series_name and not link_format else ''
          )
          for fmt in formats:
            if runstate.cancelled:
              break
            proxy.text = f'{cbook}: {fmt}'
            srcpath = cbook.formatpath(fmt)
            if srcpath is None:
              continue
            dstpath = joinpath(
                linkto_dirpath, name_subdir, name + '.' + fmt.lower()
            )
            if existspath(dstpath):
              if force:
                warning("dst already exists, will be replaced: %s", dstpath)
                (verbose or not doit) and print("unlink", shortpath(dstpath))
                doit and pfx_call(os.unlink, dstpath)
              else:
                ##warning("dst already exists, skipped: %s", dstpath)
                continue
            dstdir = dirname(dstpath)
            if not isdirpath(dstdir) and doit:
              verbose and print("mkdir", shortpath(dstdir))
              pfx_call(os.mkdir, dstdir)
            (quiet and doit
             ) or print("link", shortpath(srcpath), '=>', shortpath(dstpath))
            doit and pfx_call(os.link, srcpath, dstpath)
            if first_format:
              break
          proxy.text = f'{cbook}'
    if runstate.cancelled:
      return 1
    return 0

  # pylint: disable=too-many-locals
  def cmd_ls(self, argv):
    ''' Usage: {cmd} [-l] [-o ls-format] [book_specs...]
          List the contents of the Calibre library.
          -l            Long mode, listing book details over several lines.
          -o ls_format  Output format for use in a single line book listing.
          -r            Reverse the listing order.
          -t            Order listing by timestamp.
    '''
    options = self.options
    options.longmode = False  # pylint: disable=attribute-defined-outside-init
    options.ls_format = None
    options.sort_reverse = False
    options.sort_timestamp = False
    options.popopts(
        argv,
        l='longmode',
        o_='ls_format',
        r='sort_reverse',
        t='sort_timestamp',
    )
    if options.sort_timestamp:
      cbook_sort_key = lambda cbook: cbook.timestamp
    else:
      cbook_sort_key = self.cbook_default_sortkey
    longmode = options.longmode
    ls_format = options.ls_format
    calibre = options.calibre
    verbose = options.verbose
    upd = options.upd
    xit = 0
    cbooks = []
    if argv:
      try:
        cbooks = self.popbooks(
            argv, sortkey=cbook_sort_key, reverse=options.sort_reverse
        )
      except ValueError as e:
        raise GetoptError("invalid book specifiers: %s") from e
    else:
      with contextif(verbose, upd.run_task, "sort calibre contents"):
        calibre.preload()
        cbooks = sorted(
            calibre, key=cbook_sort_key, reverse=options.sort_reverse
        )
    runstate = options.runstate
    for cbook in cbooks:
      if runstate.cancelled:
        break
      with Pfx(cbook):
        if ls_format is None:
          top_row = []
          series_name = cbook.series_name
          if series_name:
            top_row.append(f"{series_name} [{intif(cbook.series_index)}]")
          top_row.append(cbook.title)
          author_names = cbook.author_names
          if author_names:
            top_row.extend(
                ("by", ", ".join(sorted(cbook.author_names, key=str.lower)))
            )
          top_row.append(f"({cbook.dbid})")
          if not longmode:
            top_row.append(",".join(sorted(map(str.upper, cbook.formats))))
            top_row.append(",".join(sorted(map(str.lower, cbook.tags))))
          print(*top_row)
        else:
          try:
            output = cbook.format_as(ls_format, error_sep='\n  ')
          except FormatAsError as e:
            error(str(e))
            xit = 1
            continue
          print(output)
        if longmode:
          print(" ", cbook.path)
          tags = cbook.tags
          if tags:
            print("   ", ", ".join(sorted(tags)))
          identifiers = cbook.identifiers
          if identifiers:
            print("   ", TagSet(identifiers))
          for fmt, subpath in cbook.formats.items():
            with Pfx(fmt):
              fspath = cbook.pathto(subpath)
              try:
                size = pfx_call(os.stat, fspath).st_size
              except OSError as e:
                warning("cannot stat: %s", e)
              else:
                print(f"    {fmt:4s}", transcribe_bytes_geek(size), subpath)
    if runstate.cancelled:
      xit = 1
    return xit

  def cmd_make_cbz(self, argv):
    ''' Usage: {cmd} book_specs...
          Add the CBZ format to the designated Calibre books.
    '''
    if not argv:
      raise GetoptError("missing book_specs")
    options = self.options
    runstate = options.runstate
    xit = 0
    while argv and not runstate.cancelled:
      with Pfx(argv[0]):
        try:
          cbooks = self.popbooks(argv, once=True)
        except ValueError as e:
          warning("invalid book specifications: %s", e)
          xit = 2
          continue
        for cbook in cbooks:
          if runstate.cancelled:
            break
          with Pfx(cbook):
            try:
              pfx_call(cbook.make_cbz)
            except ValueError as e:
              warning("cannot make CBZ from %s: %s" % (cbook, e))
              xit = 1
    if runstate.cancelled:
      xit = 1
    return xit

  # pylint: disable=too-many-branches
  def cmd_prefs(self, argv):
    ''' Usage: {cmd}
          List the library preferences.
    '''
    if argv:
      raise GetoptError("extra arguments: %r" % (argv,))
    xit = 0
    db = self.options.calibre.db
    with db.session() as session:
      if argv:
        for pref_name in argv:
          with Pfx(pref_name):
            pref = db.preferences.lookup1(key=pref_name, session=session)
            if pref is None:
              warning("unknown preference")
              xit = 1
            else:
              print(pref_name)
              print(" ", json.dumps(pfx_call(json.loads, pref.value)))
      else:
        for pref in sorted(db.preferences.lookup(session=session),
                           key=lambda pref: pref.key):
          with Pfx(pref.key):
            print(pref.key)
            value = pfx_call(json.loads, pref.value)
            if isinstance(value, list):
              if value:
                for item in value:
                  print(" ", json.dumps(item))
            elif isinstance(value, dict):
              for k, v in sorted(value.items()):
                print(" ", json.dumps(k), ":", json.dumps(v))
            else:
              print(" ", json.dumps(value))
    return xit

  # pylint: disable=too-many-branches,too-many-locals,too-many-statements
  def cmd_pull(self, argv):
    ''' Usage: {cmd} [-fnqv] [/path/to/other-library] [identifiers...]
          Import formats from another Calibre library.
          -f    Force. Overwrite existing formats with formats from other-library.
          -n    No action: recite planned actions.
          -q    Quiet. Only issue warnings and errors.
          -v    Verbose. Print more information.
          /path/to/other-library: optional path to another Calibre library tree
          identifier-name: the key on which to link matching books;
            the default is {DEFAULT_LINK_IDENTIFIER}
            If the identifier '?' is specified the available
            identifiers in use in other-library are listed.
          identifier-values: specific book identifiers to import
            If no identifiers are provided, all books which have
            the specified identifier will be pulled.
    '''
    options = self.options
    calibre = options.calibre
    runstate = options.runstate
    self.popopts(argv, options, f='force', n='-doit', q='quiet', v='verbose')
    if argv and argv[0].startswith('/') and isdirpath(argv[0]):
      options.calibre_path_other = argv.pop(0)
    doit = options.doit
    force = options.force
    quiet = options.quiet
    verbose = options.verbose
    upd = options.upd
    other_library = options.calibre_other
    quiet or print("pull", other_library.shortpath, "=>", calibre.shortpath)
    upd.out("pull " + shlex.join(argv))
    with Pfx(other_library.shortpath):
      with other_library:
        if other_library is calibre:
          raise GetoptError("cannot import from the same library")
        if argv:
          identifier_name = argv.pop(0)
        else:
          identifier_name = self.DEFAULT_LINK_IDENTIFIER
        if identifier_name == '?':
          if argv:
            warning(
                "ignoring extra arguments after identifier-name=?: %r", argv
            )
          print("Default identifier:", self.DEFAULT_LINK_IDENTIFIER)
          print("Available idenitifiers in %s:" % (other_library,))
          for identifier_name in sorted(other_library.identifier_names()):
            print(" ", identifier_name)
          return 0
        with upd.run_task(f'scan identifiers from {other_library}...'):
          obooks_map = {
              idv: obook
              for idv, obook in (
                  (obook.identifiers.get(identifier_name), obook)
                  for obook in other_library
              )
              if idv is not None
          }
        if not obooks_map:
          raise GetoptError(
              "no books have the identifier %r; identifiers in use are: %s" % (
                  identifier_name,
                  ', '.join(sorted(other_library.identifier_names()))
              )
          )
        if argv:
          identifier_values = argv
        else:
          identifier_values = [
              idv for idv, obook in sorted(
                  obooks_map.items(),
                  key=lambda id_ob:
                  (id_ob[1].title, id_ob[1].author_sort, id_ob[1].dbid)
              )
          ]
        xit = 0
        calibre.preload()
        with UpdProxy(prefix="pull " + other_library.shortpath + ": "
                      ) as proxy:
          for identifier_value in progressbar(identifier_values, "pull " +
                                              other_library.shortpath):
            if runstate.cancelled:
              break
            with Pfx.scope("%s=%s", identifier_name, identifier_value):
              try:
                obook = obooks_map[identifier_value]
              except KeyError:
                warning("unknown")
                xit = 1
                continue
              with proxy.extend_prefix(
                  "%s=%s: %s" % (identifier_name, identifier_value, obook)):
                if not obook.formats:
                  # pylint: disable=expression-not-assigned
                  verbose and print("no formats to pull")
                  continue
                cbooks = list(
                    calibre.by_identifier(identifier_name, identifier_value)
                )
                if not cbooks:
                  # new book
                  fmtk = list(obook.formats.keys())[0]
                  ofmtpath = obook.formatpath(fmtk)
                  # pylint: disable=expression-not-assigned
                  quiet or (
                      print(
                          "new book from %s:%s <= %s" %
                          (fmtk, obook, shortpath(ofmtpath))
                      ) if verbose else
                      print("new book from %s:%s" % (fmtk, obook))
                  )
                  dbid = calibre.add(ofmtpath, doit=doit, quiet=quiet)
                  if not doit:
                    # we didn't make a new book, so move to the next one
                    continue
                  if dbid is None:
                    error("calibre add failed")
                    xit = 1
                  else:
                    cbook = calibre[dbid]
                    # pylint: disable=expression-not-assigned
                    quiet or print('new', cbook, '<=', obook)
                elif len(cbooks) > 1:
                  # pylint: disable=expression-not-assigned
                  verbose or warning(
                      "  \n".join(
                          [
                              "multiple \"local\" books with this identifier:",
                              *map(str, cbooks)
                          ]
                      )
                  )
                  continue
                else:
                  cbook, = cbooks
                cbook.pull(
                    obook,
                    doit=doit,
                    force=force,
                    quiet=quiet,
                    verbose=verbose
                )
        if runstate.cancelled:
          xit = 1
        return xit

  def cmd_shell(self, argv):
    ''' Usage: {cmd}
          Run an interactive Python prompt with some predefined names:
          calibre: the CalibreTree
          options: self.options
    '''
    if argv:
      raise GetoptError("extra arguments: %r" % (argv,))
    options = self.options
    interact(
        banner=f'{self.cmd}: {options.calibre}',
        local=dict(
            calibre=options.calibre,
            options=options,
        )
    )

  def cmd_tag(self, argv):
    ''' Usage: {cmd} [-n] [--] [-]tag[,tag...] book_specs...
    '''
    options = self.options
    if argv and argv[0] == '-n':
      argv.pop(0)
      options.doit = False
    doit = options.doit
    upd = options.upd
    tags = self.poparg(argv, "tags")
    add_mode = True
    if tags.startswith('-'):
      add_mode = False
      tags = tags[1:]
    tag_names = sorted(map(str.lower, filter(None, tags.split(','))))
    if not tag_names:
      raise GetoptError("no tags specified")
    if not argv:
      raise GetoptError("missing book_specs")
    cbooks = self.popbooks(argv)
    with upd.insert(1) as proxy:
      for cbook in cbooks:
        proxy.text = f'{cbook} {cbook.tags}'
        tags = set(cbook.tags)
        new_tags = set(cbook.tags)
        for tag_name in tag_names:
          if add_mode:
            new_tags.add(tag_name)
          else:
            new_tags.discard(tag_name)
        if new_tags != tags:
          if add_mode:
            print(cbook, '+', sorted(new_tags - tags))
          else:
            print(cbook, '-', sorted(tags - new_tags))
          if doit:
            cbook.tags = new_tags

if __name__ == '__main__':
  sys.exit(CalibreCommand(sys.argv).run())



#!/usr/bin/env python3

''' Apple Books access.
'''

from contextlib import contextmanager
from dataclasses import dataclass
from getopt import GetoptError
from glob import glob
from os.path import join as joinpath
from pprint import pprint
import sys
from typing import Optional

from sqlalchemy import (
    Column,
    Float,
    Integer,
    LargeBinary,
    String,
)
from typeguard import typechecked

from cs.app.osx.plist import ingest_plist
from cs.cmdutils import BaseCommand
from cs.context import stackattrs
from cs.deco import strable
from cs.fs import FSPathBasedSingleton
from cs.logutils import warning
from cs.pfx import pfx_call, pfx_method
from cs.psutils import run
from cs.resources import MultiOpenMixin
from cs.sqlalchemy_utils import ORM, BasicTableMixin
from cs.threads import locked_property

class AppleBooksTree(FSPathBasedSingleton, MultiOpenMixin):
  ''' Work with an Apple Books tree.
  '''

  FSPATH_DEFAULT = '~/Library/Containers/com.apple.iBooksX/Data//Documents/BKLibrary'
  FSPATH_ENVVAR = 'APPLE_BOOKS_LIBRARY'

  @contextmanager
  def startup_shutdown(self):
    ''' For MultiOpenMixin.
    '''
    yield

  @locked_property
  def db(self):
    ''' The associated `AppleBooksDB` ORM,
        instantiated on demand.
    '''
    return AppleBooksDB(self)

  def dbshell(self):
    ''' Interactive db shell.
    '''
    return self.db.shell()

  def __iter__(self):
    ''' Generator yielding Apple Book db entries.
    '''
    db = self.db
    with db.db_session() as session:
      yield from sorted(
          db.books.lookup(session=session),
          key=lambda book: (book.sort_author, book.sort_title)
      )

class AppleBooksDB(ORM):
  ''' An ORM to access the Apple Books SQLite database.
  '''

  # example: BKLibrary-1-091020131601.sqlite
  DB_GLOB = 'BKLibrary-1-*.sqlite'

  @strable(open_func=AppleBooksTree)
  @typechecked
  def __init__(self, tree: AppleBooksTree):
    self.tree = tree
    self.db_path = self._find_library_dbpath()
    self.db_url = 'sqlite:///' + self.db_path
    super().__init__(self.db_url)

  @pfx_method
  def _find_library_dbpath(self):
    ''' Look up the path of the SQLite database.
    '''
    dbpaths = pfx_call(glob, joinpath(self.tree.fspath, self.DB_GLOB))
    if not dbpaths:
      raise ValueError("no matching library file")
    if len(dbpaths) > 1:
      dbpaths = sorted(dbpaths)
      warning(
          "  \n".join(["multiple matches, choosing the latest:", *dbpaths])
      )
      dbpath = dbpaths[-1]
    else:
      dbpath, = dbpaths
    return dbpath

  def shell(self):
    ''' Interactive db shell.
    '''
    print("sqlite3", self.db_path)
    run(['sqlite3', self.db_path], check=True)
    return 0

  @property
  def orm(self):
    ''' No distinct ORM class for `CalibreMetadataDB`.
    '''
    return self

  # lifted from SQLTags
  @contextmanager
  def db_session(self, *, new=False):
    ''' Context manager to obtain a db session if required
        (or if `new` is true).
    '''
    orm_state = self.orm.sqla_state
    get_session = orm_state.new_session if new else orm_state.auto_session
    with get_session() as session2:
      yield session2

  def declare_schema(self):
    r''' Define the database schema / ORM mapping.

        Database schema queried thus:

            echo .schema | sqlite3 /Users/cameron/Library/Containers/com.apple.iBooksX/Data//Documents/BKLibrary/BKLibrary-1-091020131601.sqlite | sed 's/\([(,]\) /\1\n  /g'
    '''
    Base = self.Base

    # pylint: disable=too-few-public-methods
    class HasPKMixin:
      ''' Include a "Z_PK" `Column` as the primary key.
      '''
      pk = Column('Z_PK', Integer, primary_key=True)

    class LibraryAsset(Base, BasicTableMixin, HasPKMixin):
      ''' A library asset, presumeably a book.

          This has many columns, I'm just defining the interesting ones.
          We get away with this because we're only selecting, not inserting.
      '''
      __tablename__ = 'ZBKLIBRARYASSET'
      content_type = Column('ZCONTENTTYPE', Integer)
      file_size = Column('ZFILESIZE', Integer)
      page_size = Column('ZPAGECOUNT', Integer)
      creation_date = Column('ZCREATIONDATE', Float)  # not a DateTime
      purchase_date = Column('ZPURCHASEDATE', Float)  # not a DateTime
      release_date = Column('ZRELEASEDATE', Float)  # not a DateTime
      version_number = Column('ZVERSIONNUMBER', Float)
      account_id = Column('ZACCOUNTID', String)
      asset_id = Column('ZASSETID', String)
      author = Column('ZAUTHOR', String)
      description = Column('ZBOOKDESCRIPTION', String)
      cover_url = Column('ZCOVERURL', String)
      epub_id = Column('ZEPUBID', String)
      family_id = Column('ZFAMILYID', String)
      genre = Column('ZGENRE', String)
      grouping = Column('ZGROUPING', String)
      kind = Column('ZKIND', String)
      language = Column('ZLANGUAGE', String)
      path = Column('ZPATH', String)
      perm_link = Column('ZPERMLINK', String)
      series_id = Column('ZSERIESID', String)
      sort_author = Column('ZSORTAUTHOR', String)
      sort_title = Column('ZSORTTITLE', String)
      store_id = Column('ZSTOREID', String)
      title = Column('ZTITLE', String)
      human_version_number = Column('ZVERSIONNUMBERHUMANREADABLE', String)
      year = Column('ZYEAR', String)

    class Metadata(Base, BasicTableMixin):
      ''' The metadata table - seems to be a single row with library metadata in a plist.
      '''
      __tablename__ = 'Z_METADATA'
      version = Column('Z_VERSION', Integer, primary_key=True)
      uuid = Column('Z_UUID', String)
      plist = Column('Z_PLIST', LargeBinary)

      def plist_as_dict(self):
        ''' Return the plist column as a `PListDict`.
        '''
        return ingest_plist(self.plist)

    class Collection(Base, BasicTableMixin, HasPKMixin):
      ''' Collection definitions.
      '''
      __tablename__ = 'ZBKCOLLECTION'
      collection_id = Column('ZCOLLECTIONID', String)
      details = Column('ZDETAILS', String)
      title = Column('ZTITLE', String)

    class CollectionMember(Base, BasicTableMixin, HasPKMixin):
      ''' Collection members.
      '''
      __tablename__ = 'ZBKCOLLECTIONMEMBER'
      asset = Column('ZASSET', Integer)
      collection = Column('ZCOLLECTION', Integer)
      asset_id = Column('ZASSETID', String)

    self.books = LibraryAsset
    self.metadata = Metadata
    self.collections = Collection
    self.collection_members = CollectionMember

class AppleBooksCommand(BaseCommand):
  ''' Command line access to Apple Books.
  '''

  @dataclass
  class Options(BaseCommand.Options):
    apple_path: Optional[str] = None

  @contextmanager
  def run_context(self):
    ''' Prepare the `SQLTags` around each command invocation.
    '''
    with super().run_context():
      options = self.options
      with AppleBooksTree(options.apple_path) as at:
        with stackattrs(options, apple=at, verbose=True):
          yield

  def cmd_dbshell(self, argv):
    ''' Usage: {cmd}
          Start an interactive database shell.
    '''
    if argv:
      raise GetoptError("extra arguments: %r" % (argv,))
    return self.options.apple.dbshell()

  def cmd_ls(self, argv):
    ''' Usage: {cmd}
          List books in the library.
    '''
    options = self.options
    long_mode = False
    if argv and argv[0] == '-l':
      long_mode = True
      argv.pop(0)
    if argv:
      raise GetoptError("extra arguments: %r" % (argv,))
    for book in options.apple:
      print(f"{book.title}, {book.author} ({book.asset_id})")
      if long_mode:
        if book.cover_url:
          print("  cover:", book.cover_url)
        if book.genre:
          print("  genre:", book.genre)
        if book.description:
          print("  description:")
          print("   ", book.description.replace("\n", "    \n"))

  def cmd_md(self, argv):
    ''' Usage: {cmd}
          List metadata.
    '''
    options = self.options
    if argv:
      raise GetoptError("extra arguments: %r" % (argv,))
    db = options.apple.db
    with db.db_session() as session:
      for md in db.metadata.lookup(session=session):
        pprint(md.plist_as_dict())

if __name__ == '__main__':
  sys.exit(AppleBooksCommand(sys.argv).run())



#!/usr/bin/env python3

''' Support for MOBI format ebook files.
'''

from contextlib import contextmanager
from getopt import GetoptError
from glob import glob
import os
from os.path import (
    basename,
    exists as existspath,
    isfile as isfilepath,
    join as joinpath,
    relpath,
    splitext,
)
import sys
from tempfile import TemporaryDirectory
from zipfile import ZipFile, ZIP_STORED

import mobi

from cs.cmdutils import BaseCommand
from cs.context import stackattrs
from cs.logutils import error, info
from cs.pfx import pfx, pfx_call

class Mobi:
  ''' Work with an existing MOBI ebook file.
  '''

  def __init__(self, mobipath):
    if not isfilepath(mobipath):
      raise ValueError("mobipath %r is not a file" % (mobipath,))
    self.path = mobipath

  def __repr__(self):
    return "%s(%r)" % (type(self).__name__, repr(self.path))

  __str__ = __repr__

  @pfx
  def extract(self, dirpath=None):
    ''' Extract the contents of the MOBI file into a directory.
        Return `(dirpath,rfilepath)` where `dirpath` is the extracted
        file tree and `filepath` is the relative pathname of the
        primary epub, html or pdf file depending on the mobi type.
    '''
    if dirpath is not None and existspath(dirpath):
      raise ValueError("dirpath %r already exists" % (dirpath,))
    # divert stdout because the mobi library sends some warnings etc to stdout
    with stackattrs(sys, stdout=sys.stderr):
      tmpdirpath, filepath = pfx_call(mobi.extract, self.path)
    rfilepath = relpath(filepath, tmpdirpath)
    if dirpath is None:
      dirpath = tmpdirpath
    else:
      pfx_call(os.rename, tmpdirpath, dirpath)
    return dirpath, rfilepath

  @contextmanager
  def extracted(self):
    ''' Context manager version of `extract()` which extracts the
        MOBI into a temporary directory and yields the resulting
        `(dirpath,rfilepath)` as for `extract()`.
    '''
    with TemporaryDirectory(prefix='%s.extracted-' % (type(self).__name__,),
                            suffix='-%s' %
                            (self.path.replace(os.sep, '_'),)) as T:
      dirpath, rfilepath = self.extract(dirpath=joinpath(T, 'extracted'))
      yield dirpath, rfilepath

  @pfx
  def make_cbz(self, cbzpath=None):
    ''' Create a CBZ file from the images in the MOBI file.
        Return the path to the created CBZ file.
    '''
    if cbzpath is None:
      mobibase, _ = splitext(basename(self.path))
      cbzpath = mobibase + '.cbz'
    if existspath(cbzpath):
      raise ValueError("CBZ path %r already exists" % (cbzpath,))
    with self.extracted() as df:
      dirpath, _ = df
      imagepaths = sorted(glob(joinpath(dirpath, 'mobi8/OEBPS/Images/*.*')))
      if not imagepaths:
        imagepaths = sorted(glob(joinpath(dirpath, 'mobi7/Images/*.*')))
      if not imagepaths:
        for dirp, dirnames, filenames in os.walk(dirpath):
          dirnames[:] = sorted(dirnames)
          for f in sorted(filenames):
            print(joinpath(dirp, f))
        raise ValueError("no image paths")
      info("write %s", cbzpath)
      try:
        with pfx_call(ZipFile, cbzpath, 'x', compression=ZIP_STORED) as cbz:
          for imagepath in imagepaths:
            pfx_call(cbz.write, imagepath, arcname=basename(imagepath))
      except FileExistsError as e:
        error("CBZ already eixsts: %r: %s", cbzpath, e)
        return 1
      except Exception:
        if existspath(cbzpath):
          pfx_call(os.unlink, cbzpath)
        raise
    return cbzpath

class MobiCommand(BaseCommand):
  ''' Command line implementation for `mobi2cbz`.
  '''

  def cmd_extract(self, argv):
    ''' Usage: {cmd} mobipath [outdir]
          Extract the contents of the MOBI file mobipath
          into the directory outdir, default based on the mobipath basename.
          Prints the outdir and the name of the top file.
    '''
    outdirpath = None
    mobipath = self.poparg(argv, "mobipath")
    if argv:
      outdirpath = argv.pop(0)
    if argv:
      raise GetoptError("extra arguments after outdir: %r" % (argv,))
    if not existspath(mobipath):
      raise GetoptError("mobipath does not exist: %r" % (mobipath,))
    if outdirpath is None:
      outdirpath, _ = splitext(basename(mobipath))
    if existspath(outdirpath):
      raise GetoptError("outdir already exists: %s" % (outdirpath,))
    MB = Mobi(mobipath)
    extdirpath, rfilepath = MB.extract(outdirpath)
    assert extdirpath == outdirpath
    print(outdirpath)
    print(rfilepath)

  def cmd_make_cbz(self, argv):
    ''' Usage: {cmd} mobipath [cbzpath]
          Unpack a MOBI file and construct a CBZ file.
          Prints the path of the CBZ file to the output.
          The default cbzpath is mobibase.cbz where mobibase is the
          basename of mobipath with its extension removed.
    '''
    mobipath = self.poparg(argv, "mobipath")
    mobibase, _ = splitext(basename(mobipath))
    if argv:
      cbzpath = argv.pop(0)
    else:
      cbzpath = mobibase + '.cbz'
    if argv:
      raise GetoptError("extra arguments after cbzpath: %r" % (argv,))
    if not existspath(mobipath):
      raise GetoptError("mobipath does not exist: %r" % (mobipath,))
    if existspath(cbzpath):
      raise GetoptError("CBZ already exists: %r" % (cbzpath,))
    MB = Mobi(mobipath)
    outcbzpath = MB.make_cbz(cbzpath)
    print(outcbzpath)



#!/usr/bin/env python3

''' Utilities and command line for working with EBooks.
    Basic support for talking to Apple Books, Calibre, Kindle, Mobi.
    These form the basis of my personal Kindle and Calibre workflow.

    The command `python -m cs.ebooks help -l` gives the basic usage information:

        help:
          apple subcommand [...]
              Subcommands:
                dbshell
                  Start an interactive database shell.
                help [-l] [subcommand-names...]
                  Print the full help for the named subcommands,
                  or for all subcommands if no names are specified.
                  -l  Long help even if no subcommand-names provided.
                ls
                  List books in the library.
                md
                  List metadata.
                shell
                  Run a command prompt via cmd.Cmd using this command's subcommands.
          calibre [-C calibre_library] [-K kindle-library-path] subcommand [...]
              -C calibre_library
                Specify calibre library location.
              -K kindle_library
                Specify kindle library location.
              -O other_calibre_library
                Specify alternate calibre library location, the default library
                for pull etc. The default comes from $CALIBRE_LIBRARY_OTHER.
              Subcommands:
                add [-nqv] bookpaths...
                  Add the specified ebook bookpaths to the library.
                  -n    No action: recite planned actions.
                  -q    Quiet: only emit warnings.
                  -v    Verbose: report all actions and decisions.
                convert [-fnqv] formatkey dbids...
                  Convert books to the format `formatkey`.
                  -f    Force: convert even if the format is already present.
                  -n    No action: recite planned actions.
                  -q    Quiet: only emit warnings.
                  -v    Verbose: report all actions and decisions.
                dbshell
                  Start an interactive database prompt.
                help [-l] [subcommand-names...]
                  Print the full help for the named subcommands,
                  or for all subcommands if no names are specified.
                  -l  Long help even if no subcommand-names provided.
                info
                  Report basic information.
                linkto [-1fnqv] [-d linkto-dir] [-F fmt,...] [-o link-format] [dbids...]
                  Export books to linkto-dir by hard linking.
                  -1              Link only the first format found.
                  -d linkto-dir   Specify the target directory, default from $MEDIA_BOOKSPATH
                                  or ~/media/books.
                  -F fmt,...      Source formats, default: ['CBZ', 'EPUB']
                  -f              Force. Replace existing links.
                  -n              No action. Report planned actions.
                  -o link-format  Link name format.
                  -q              Quiet.
                  -v              Verbose.
                ls [-l] [-o ls-format] [book_specs...]
                  List the contents of the Calibre library.
                  -l            Long mode, listing book details over several lines.
                  -o ls_format  Output format for use in a single line book listing.
                  -r            Reverse the listing order.
                  -t            Order listing by timestamp.
                make_cbz book_specs...
                  Add the CBZ format to the designated Calibre books.
                prefs
                  List the library preferences.
                pull [-fnqv] [/path/to/other-library] [identifiers...]
                  Import formats from another Calibre library.
                  -f    Force. Overwrite existing formats with formats from other-library.
                  -n    No action: recite planned actions.
                  -q    Quiet. Only issue warnings and errors.
                  -v    Verbose. Print more information.
                  /path/to/other-library: optional path to another Calibre library tree
                  identifier-name: the key on which to link matching books;
                    the default is mobi-asin
                    If the identifier '?' is specified the available
                    identifiers in use in other-library are listed.
                  identifier-values: specific book identifiers to import
                    If no identifiers are provided, all books which have
                    the specified identifier will be pulled.
                shell
                  Run an interactive Python prompt with some predefined names:
                  calibre: the CalibreTree
                  options: self.options
                tag [-n] [--] [-]tag[,tag...] book_specs...
          dedrm [-D dedrm_package_path] subcommand [args...]
                -D  Specify the filesystem path to the DeDRM/noDRM plugin top level.
                    For example, if you had a checkout of git@github.com:noDRM/DeDRM_tools.git
                    at /path/to/DeDRM_tools--noDRM you could supply:
                    -D /path/to/DeDRM_tools--noDRM/DeDRM_plugin
                    or place that value in the $DEDRM_PACKAGE_PATH environment variable.
              Subcommands:
                help [-l] [subcommand-names...]
                  Print the full help for the named subcommands,
                  or for all subcommands if no names are specified.
                  -l  Long help even if no subcommand-names provided.
                import module_name...
                  Exercise the DeDRM python import mechanism for each module_name.
                kindlekeys [import]
                  import    Read a JSON list of key dicts and update the cached keys.
                remove filenames...
                  Remove DRM from the specified filenames.
                shell
                  Run a command prompt via cmd.Cmd using this command's subcommands.
          help [-l] [subcommand-names...]
            Print the full help for the named subcommands,
            or for all subcommands if no names are specified.
            -l  Long help even if no subcommand-names provided.
          kindle [-C calibre_library] [-K kindle-library-path] [subcommand [...]]
              -C calibre_library
                Specify calibre library location.
              -K kindle_library
                Specify kindle library location.
              Subcommands:
                app_path [content-path]
                  Report or set the content path for the Kindle application.
                dbshell
                  Start an interactive database prompt.
                export [-fnqv] [ASINs...]
                  Export AZW files to Calibre library.
                  -f    Force: replace the AZW3 format if already present.
                  -n    No action, recite planned actions.
                  -q    Quiet: report only warnings.
                  -v    Verbose: report more information about actions and inaction.
                  ASINs Optional ASIN identifiers to export.
                        The default is to export all books with no "calibre.dbid" fstag.
                help [-l] [subcommand-names...]
                  Print the full help for the named subcommands,
                  or for all subcommands if no names are specified.
                  -l  Long help even if no subcommand-names provided.
                import_tags [-nqv] [ASINs...]
                  Import Calibre book information into the fstags for a Kindle book.
                  This will support doing searches based on stuff like
                  titles which are, naturally, not presented in the Kindle
                  metadata db.
                info
                  Report basic information.
                ls [-l]
                  List the contents of the library.
                  -l  Long mode.
                shell
                  Run a command prompt via cmd.Cmd using this command's subcommands.
          mobi subcommand [...]
              Subcommands:
                extract mobipath [outdir]
                  Extract the contents of the MOBI file mobipath
                  into the directory outdir, default based on the mobipath basename.
                  Prints the outdir and the name of the top file.
                help [-l] [subcommand-names...]
                  Print the full help for the named subcommands,
                  or for all subcommands if no names are specified.
                  -l  Long help even if no subcommand-names provided.
                make_cbz mobipath [cbzpath]
                  Unpack a MOBI file and construct a CBZ file.
                  Prints the path of the CBZ file to the output.
                  The default cbzpath is mobibase.cbz where mobibase is the
                  basename of mobipath with its extension removed.
                shell
                  Run a command prompt via cmd.Cmd using this command's subcommands.
          shell
            Run a command prompt via cmd.Cmd using this command's subcommands.
'''

__version__ = '20230704'

DISTINFO = {
    'keywords': ["python3"],
    'classifiers': [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    'install_requires': [
        'cs.app.osx.defaults',
        'cs.app.osx.plist',
        'cs.cmdutils',
        'cs.context',
        'cs.deco',
        'cs.fileutils',
        'cs.fs',
        'cs.fstags',
        'cs.lex',
        'cs.logutils',
        'cs.numeric',
        'cs.obj',
        'cs.pfx',
        'cs.progress',
        'cs.psutils',
        'cs.resources',
        'cs.sqlalchemy_utils',
        'cs.sqltags',
        'cs.tagset',
        'cs.threads',
        'cs.units',
        'cs.upd',
        'icontract',
        'mobi',
        'pycryptodomex',
        'sqlalchemy',
        'typeguard',
    ],
}



#!/usr/bin/env python3

''' Support for Kindle libraries.
'''

from contextlib import contextmanager
from dataclasses import dataclass, field
import filecmp
from getopt import GetoptError
import os
from os.path import (
    dirname,
    expanduser,
    exists as existspath,
    isdir as isdirpath,
    isfile as isfilepath,
    join as joinpath,
)
import sys
from typing import Optional

from icontract import require
from sqlalchemy import (
    Column,
    ForeignKey,
    Integer,
    String,
)
from sqlalchemy.orm import relationship
from typeguard import typechecked
try:
  from lxml import etree
except ImportError:
  import xml.etree.ElementTree as etree

from cs.app.osx.defaults import DomainDefaults as OSXDomainDefaults
from cs.cmdutils import BaseCommand
from cs.context import stackattrs
from cs.deco import cachedmethod, fmtdoc
from cs.fileutils import shortpath
from cs.fs import FSPathBasedSingleton, HasFSPath
from cs.fstags import FSTags
from cs.lex import cutsuffix
from cs.logutils import warning
from cs.obj import SingletonMixin
from cs.pfx import Pfx, pfx_call
from cs.progress import progressbar
from cs.psutils import run
from cs.resources import MultiOpenMixin
from cs.sqlalchemy_utils import (
    ORM,
    BasicTableMixin,
    HasIdMixin,
    RelationProxy,
)
from cs.upd import Upd, print  # pylint: disable=redefined-builtin

from .dedrm import DeDRMWrapper, DEDRM_PACKAGE_PATH_ENVVAR

def main(argv=None):
  ''' Kindle command line mode.
  '''
  return KindleCommand(argv).run()

KINDLE_LIBRARY_ENVVAR = 'KINDLE_LIBRARY'

KINDLE_APP_OSX_DEFAULTS_DOMAIN = 'com.kindle.Kindle'
KINDLE_APP_OSX_DEFAULTS_CONTENT_PATH_SETTING = 'User Settings.CONTENT_PATH'
KINDLE_APP_OSX_DEFAULTS_CONTENT_PATH = (
    '~/Library/Containers/com.amazon.Kindle/Data/Library/'
    'Application Support/Kindle/My Kindle Content'
)

def kindle_content_path_default():
  ''' Return the default content path for the Kindle application.
      Currently only knows about Darwin (MacOS).
  '''
  if sys.platform == 'darwin':
    return expanduser(KINDLE_APP_OSX_DEFAULTS_CONTENT_PATH)
  raise RuntimeError(
      "I do not know the default Kindle content path on platform %r" %
      (sys.platform,)
  )

def kindle_content_path():
  ''' Return the default Kindle content path or `None`.
      Currently only supports Darwin (MacOS).
  '''
  if sys.platform == 'darwin':
    defaults = OSXDomainDefaults(KINDLE_APP_OSX_DEFAULTS_DOMAIN)
    path = defaults.get(KINDLE_APP_OSX_DEFAULTS_CONTENT_PATH_SETTING)
    if path is None:
      path = kindle_content_path_default()
    return path
  raise RuntimeError(
      "cannot look up Kindle content path on platform %r" % (sys.platform,)
  )

@fmtdoc
def default_kindle_library():
  ''' Return the default kindle library content path
        from ${KINDLE_LIBRARY_ENVVAR}.
        On Darwin, fall back to the Kindle app setting from the
        defaults domain {KINDLE_APP_OSX_DEFAULTS_DOMAIN!r}
        setting {KINDLE_APP_OSX_DEFAULTS_CONTENT_PATH_SETTING!r}.
        Returns `None` if no default can be found.
    '''
  path = os.environ.get(KINDLE_LIBRARY_ENVVAR, None)
  if path is not None:
    return path
  return kindle_content_path()

class KindleTree(FSPathBasedSingleton, MultiOpenMixin):
  ''' Work with a Kindle ebook tree.

      This actually knows very little about Kindle ebooks or its rather opaque database.
      This is mostly to aid keeping track of state using `cs.fstags`.
  '''

  CONTENT_DIRNAME = 'My Kindle Content'

  FSPATH_DEFAULT = default_kindle_library()

  FSPATH_ENVVAR = KINDLE_LIBRARY_ENVVAR

  SUBDIR_SUFFIXES = '_EBOK', '_EBSP'

  def __init__(self, fspath=None):
    if hasattr(self, '_bookrefs'):
      return
    super().__init__(fspath=fspath)
    self._bookrefs = {}

    # define the proxy classes
    class KindleBook(SingletonMixin, RelationProxy(self.db.books, [
        'asin',
        'type',
        'revision',
        'sampling',
    ], id_column='asin'), HasFSPath):
      ''' A reference to a Kindle library book subdirectory.
      '''

      # pylint: disable=unused-argument
      @classmethod
      def _singleton_key(cls, tree: KindleTree, subdir_name):
        ''' The singleton key is `(tree,subdir_name
        '''
        return id(tree), subdir_name

      @typechecked
      @require(lambda subdir_name: os.sep not in subdir_name)
      def __init__(self, tree: KindleTree, subdir_name: str):
        ''' Initialise this book subdirectory reference.

            Parameters:
            * `tree`: the `Kindletree` containing the subdirectory
            * `subdir_name`: the subdirectory name
        '''
        if 'tree' not in self.__dict__:
          self.tree = tree
          self.subdir_name = subdir_name
          super().__init__(self.asin)

      def __str__(self):
        return "%s[%s]:%s" % (self.tree, self.subdir_name, self.tags)

      def __repr__(self):
        return "%s(%r,%r)" % (type(self).__name__, self.tree, self.subdir_name)

      @property
      def fspath(self):
        ''' The filesystem path of this book subdirectory.
        '''
        return self.tree.pathto(self.subdir_name)

      @property
      def asin(self):
        ''' The ASIN of this book subdirectory, normalised to upper case.
        '''
        for suffix in '_EBOK', '_EBSP':
          prefix = cutsuffix(self.subdir_name, suffix)
          if prefix is not self.subdir_name:
            return prefix.upper()
        raise ValueError(
            "subdir_name %r does not end with _EBOK or _BSP" %
            (self.subdir_name,)
        )

      @property
      def amazon_url(self):
        ''' The Amazon product page for this book.
        '''
        # https://www.amazon.com.au/Wonder-Woman-2016-xx-Liars-ebook/dp/B097KMW2VY/
        title = self.tags.get('calibre.title',
                              'title').replace(' ', '-').replace('/', '-')
        return f'https://www.amazon.com.au/{title}/dp/{self.asin}/'

      def listdir(self):
        ''' Return a list of the names inside the subdirectory,
              or an empty list if the subdirectory is not present.
          '''
        try:
          return os.listdir(self.fspath)
        except FileNotFoundError:
          return []

      def extpath(self, ext):
        ''' Return the filesystem path to the booknamed file
            within the book subdirectory.
        '''
        return self.pathto(self.subdir_name + '.' + ext)

      @property
      def tags(self):
        ''' The `FSTags` for this book subdirectory.
        '''
        return self.tree.fstags[self.fspath]

      def asset_names(self):
        ''' Return the names of files within the subdirectory
            whose names start with `self.subdir_name+'.'`.
        '''
        prefix_ = self.subdir_name
        return [name for name in self.listdir() if name.startswith(prefix_)]

      def subpath(self, name):
        ''' The filesystem path of `name` within this subdirectory.
        '''
        return joinpath(self.fspath, name)

      def phl_xml(self):
        ''' Decode the `.phl` XML file if present and return an XML `ElementTree`.
            Return `None` if the file is not present.

            This file seems to contain popular highlights in the
            `popular/content/annotation` tags.
          '''
        phl_path = self.subpath(self.subdir_name + '.phl')
        try:
          with pfx_call(open, phl_path, 'rb') as f:
            xml_bs = f.read()
        except FileNotFoundError:
          return None
        with Pfx(phl_path):
          return pfx_call(etree.fromstring, xml_bs)

      # pylint: disable=too-many-branches
      def export_to_calibre(
          self,
          calibre,
          *,
          dedrm=None,
          doit=True,
          replace_format=False,
          force=False,
          quiet=False,
          verbose=False,
      ):
        ''' Export this Kindle book to a Calibre instance,
            return `(cbook,added)`
            being the `CalibreBook` and whether the Kinble book was added
            (books are not added if the format is already present).

            Parameters:
            * `calibre`: the `CalibreTree`
            * `dedrm`: optional `DeDRMWrapper` instance
            * `doit`: optional flag, default `True`;
              if false just recite planned actions
            * `force`: optional flag, default `False`;
              if true pull the AZW file even if an AZW format already exists
            * `replace_format`: if true, export even if the `AZW3`
              format is already present
            * `quiet`: default `False`, do not print nonwarnings
            * `verbose`: default `False`, print all actions or nonactions
        '''
        azwpath = self.extpath('azw')
        if not isfilepath(azwpath):
          raise ValueError("no AZW file: %r" % (azwpath,))
        added = False
        cbooks = list(calibre.by_asin(self.asin))
        if not cbooks:
          # new book
          # pylint: disable=expression-not-assigned
          quiet or print("new book <=", shortpath(azwpath))
          dbid = calibre.add(azwpath, dedrm=dedrm, doit=doit, quiet=quiet)
          if dbid is None:
            added = not doit
            cbook = None
          else:
            added = True
            cbook = calibre[dbid]
            quiet or print(" ", cbook)
        else:
          # book already present in calibre
          cbook = cbooks[0]
          if len(cbooks) > 1:
            warning(
                "multiple calibre books, dbids %r: choosing %s",
                [cb.dbid for cb in cbooks], cbook
            )
          with Pfx(cbook):
            # look for exact content match
            for fmtk in 'AZW3', 'AZW', 'MOBI':
              fmtpath = cbook.formatpath(fmtk)
              if fmtpath and filecmp.cmp(fmtpath, azwpath):
                # pylint: disable=expression-not-assigned
                verbose and print(
                    cbook, fmtk, shortpath(fmtpath), '=', shortpath(azwpath)
                )
                return cbook, False
            # remaining logic is in CalibreBook.pull_format
            cbook.pull_format(
                azwpath, doit=doit, force=force, quiet=quiet, verbose=verbose
            )
        return cbook, added

    self.KindleBook = KindleBook

  def __str__(self):
    return "%s:%s" % (type(self).__name__, shortpath(self.fspath))

  def __repr__(self):
    return "%s(%r)" % (type(self).__name__, self.fspath)

  @contextmanager
  def startup_shutdown(self):
    ''' Context manager to obtain and release resources.
    '''
    with FSTags() as fstags:
      with stackattrs(self, fstags=fstags):
        yield

  @property
  @cachedmethod
  def db(self):
    ''' The associated `KindleBookAssetDB` ORM,
        instantiated on demand.
    '''
    return KindleBookAssetDB(self)

  def dbshell(self):
    ''' Interactive db shell.
    '''
    return self.db.shell()

  def is_book_subdir(self, subdir_name):
    ''' Test whether `subdir_name` is a Kindle ebook subdirectory basename.
    '''
    return subdir_name.endswith(self.SUBDIR_SUFFIXES)

  def book_subdir_names(self):
    ''' Return a list of the individual ebook subdirectory names.
    '''
    return [
        dirbase for dirbase in os.listdir(self.fspath)
        if self.is_book_subdir(dirbase)
    ]

  def asins(self):
    ''' Return a `set` of the `books.asin` column values.
    '''
    db = self.db
    books = db.books
    with db.session() as session:
      return set(asin for asin, in session.query(books.asin))

  def by_asin(self, asin):
    ''' Return a `KindleBook` for the supplied `asin`.
    '''
    ASIN = asin.upper()
    # convert ASIN_EBOK to ASIN, handy for copy/paste of subdir name from listing
    for suffix in self.SUBDIR_SUFFIXES:
      subdir_name = cutsuffix(ASIN, suffix)
      if subdir_name is not ASIN:
        ASIN = subdir_name
        break
    for suffix in self.SUBDIR_SUFFIXES:
      subdir_name = ASIN + suffix
      if isdirpath(self.pathto(subdir_name)):
        return self[subdir_name]
    return self[ASIN + self.SUBDIR_SUFFIXES[0]]

  def keys(self):
    ''' The keys of a `KindleTree` are its book subdirectory names.
    '''
    return self.book_subdir_names()

  def __getitem__(self, subdir_name):
    ''' Return the `cs.fstags.TaggedPath` for the ebook subdirectory named `subdir_name`.
    '''
    if not self.is_book_subdir(subdir_name):
      raise ValueError(
          "not a Kindle ebook subdirectory name: %r" % (subdir_name,)
      )
    try:
      book = self._bookrefs[subdir_name]
    except KeyError:
      book = self._bookrefs[subdir_name] = self.KindleBook(self, subdir_name)
    return book

  def __iter__(self):
    ''' Mapping iteration method.
    '''
    return iter(self.keys())

  def values(self):
    ''' Mapping method yielding `KindleBook` instances.
    '''
    yield from map(self.__getitem__, self)

  def items(self):
    ''' Mapping method yielding `(subdir_name,KindleBook)` pairs.
    '''
    for k in self:
      yield k, self[k]

# pylint: disable=too-many-instance-attributes
class KindleBookAssetDB(ORM):
  ''' An ORM to access the Kindle `book_asset.db` SQLite database.
  '''

  DB_FILENAME = 'book_asset.db'

  def __init__(self, tree):
    self.tree = tree
    self.db_url = 'sqlite:///' + self.db_path
    super().__init__(self.db_url)

  @property
  def orm(self):
    ''' No distinct ORM class for `KindleBookAssetDB`.
    '''
    return self

  @property
  def db_path(self):
    ''' The filesystem path to the database.
    '''
    return self.tree.pathto(self.DB_FILENAME)

  def shell(self):
    ''' Interactive db shell.
    '''
    print("sqlite3", self.db_path)
    with Upd().above():
      run(['sqlite3', self.db_path], check=True)
    return 0

  # lifted from SQLTags
  @contextmanager
  def session(self, *, new=False):
    ''' Context manager to obtain a db session if required
        (or if `new` is true).
    '''
    orm_state = self.orm.sqla_state
    get_session = orm_state.new_session if new else orm_state.auto_session
    with get_session() as session2:
      yield session2

  # pylint: disable=too-many-statements
  def declare_schema(self):
    r''' Define the database schema / ORM mapping.

        Database schema queried thus:

            sqlite3 ~/KINDLE/book_asset.db .schema | sed 's/,  *\([^ ]\)/,\n    \1/g'
    '''
    Base = self.Base

    # pylint: disable=missing-class-docstring
    class VersionInfo(Base, BasicTableMixin):
      __tablename__ = 'VersionInfo'
      version = Column(
          Integer,
          primary_key=True,
          comment='database schema version number, just one row'
      )

    # pylint: disable=missing-class-docstring
    class DownloadState(Base, BasicTableMixin, HasIdMixin):
      __tablename__ = 'DownloadState'
      state = Column(
          String,
          nullable=False,
          unique=True,
          comment='mapping of download state ids to text description'
      )

    # pylint: disable=missing-class-docstring
    class Book(Base, BasicTableMixin, HasIdMixin):
      __tablename__ = 'Book'
      asin = Column(String, nullable=False, comment='Amazon ASIN indentifier')
      type = Column(
          String, nullable=False, comment='Book type eg "kindle.ebook"'
      )
      revision = Column(
          String, nullable=False, comment='Book revision, often blank'
      )
      sampling = Column(
          String,
          nullable=False,
          comment=
          'Book sample state, often blank, "Sample" for a sample download'
      )

    # pylint: disable=missing-class-docstring
    class BookDownloadInfo(Base, BasicTableMixin, HasIdMixin):
      __tablename__ = 'BookDownloadInfo'
      bookId = Column(
          Integer,
          ForeignKey("Book.id"),
          nullable=False,
          index=True,
          comment='Book row id',
      )
      book = relationship("Book")
      responseContext = Column(String, comment='BASE64 encoded information')

    # pylint: disable=missing-class-docstring
    class RequirementLevel(Base, BasicTableMixin, HasIdMixin):
      __tablename__ = 'RequirementLevel'
      level = Column(
          String,
          nullable=False,
          unique=True,
          comment='mapping of requirement level ids to text description'
      )

    # pylint: disable=missing-class-docstring
    class Asset(Base, BasicTableMixin, HasIdMixin):
      __tablename__ = 'Asset'
      bookId = Column(
          Integer,
          ForeignKey("Book.id"),
          nullable=False,
          index=True,
          comment='Book row id',
      )
      book = relationship("Book")
      guid = Column(String, nullable=False, comment='GUID of the Asset?')
      requirementLevel_id = Column(
          "requirementLevel",
          Integer,
          ForeignKey("RequirementLevel.id"),
          nullable=False,
          index=True,
          comment='Requirement Level id'
      )
      requirementLevel = relationship("RequirementLevel")
      size = Column(Integer)
      contentType = Column(String, nullable=False)
      localFilename = Column(String)
      downloadState_id = Column(
          "downloadState",
          Integer,
          ForeignKey("DownloadState.id"),
          default=1,
          comment='Asset download state id'
      )
      downloadState = relationship("DownloadState")

    # pylint: disable=missing-class-docstring
    class EndpointType(Base, BasicTableMixin, HasIdMixin):
      __tablename__ = 'EndpointType'
      type = Column(String, nullable=False)

    # pylint: disable=missing-class-docstring
    class DeliveryType(Base, BasicTableMixin, HasIdMixin):
      __tablename__ = 'DeliveryType'
      type = Column(String, nullable=False)

    # pylint: disable=missing-class-docstring
    class AssetDownloadInfo(Base, BasicTableMixin, HasIdMixin):
      __tablename__ = 'AssetDownloadInfo'
      assetId = Column(
          Integer,
          ForeignKey("Asset.id"),
          nullable=False,
          index=True,
          comment='Asset row id',
      )
      endpoint = Column(String, nullable=False)
      endpointType = Column(
          Integer,
          ForeignKey("EndpointType.id"),
          nullable=False,
          index=True,
          comment='Endpoint type id'
      )
      responseContext = Column(String)
      deliveryType = Column(
          Integer,
          ForeignKey("DeliveryType.id"),
          nullable=False,
          index=True,
          comment='Delivery type id'
      )

    # just suck the version out
    with self.session() as session:
      self.version_info = VersionInfo.lookup1(session=session).version

    # references to table definitions
    self.download_state_map = DownloadState
    DownloadState.orm = self
    self.books = Book
    Book.orm = self
    self.book_download_info = BookDownloadInfo
    BookDownloadInfo.orm = self
    self.requirement_level_map = RequirementLevel
    RequirementLevel.orm = self
    self.assets = Asset
    Asset.orm = self
    self.endpoint_type_map = EndpointType
    EndpointType.orm = self
    self.delivery_type_map = DeliveryType
    DeliveryType.orm = self
    self.asset_download_info = AssetDownloadInfo
    AssetDownloadInfo.orm = self

class KindleCommand(BaseCommand):
  ''' Command line for interacting with a Kindle filesystem tree.
  '''

  GETOPT_SPEC = 'C:K:'

  USAGE_FORMAT = '''Usage: {cmd} [-C calibre_library] [-K kindle-library-path] [subcommand [...]]
  -C calibre_library
    Specify calibre library location.
  -K kindle_library
    Specify kindle library location.'''

  SUBCOMMAND_ARGV_DEFAULT = 'info'

  @dataclass
  class Options(BaseCommand.Options):
    ''' Set up the default values in `options`.
    '''

    def _kindle_path():
      try:
        # pylint: disable=protected-access
        kindle_path = KindleTree._resolve_fspath(None)
      except ValueError:
        kindle_path = None
      return kindle_path

    kindle_path: Optional[str] = field(default_factory=_kindle_path)

    def _calibre_path():
      from .calibre import CalibreTree  # pylint: disable=import-outside-toplevel
      try:
        # pylint: disable=protected-access
        calibre_path = CalibreTree._resolve_fspath(None)
      except ValueError:
        calibre_path = None
      return calibre_path

    calibre_path: Optional[str] = field(default_factory=_calibre_path)
    dedrm_package_path: Optional[str] = field(
        default_factory=lambda: os.environ.get(DEDRM_PACKAGE_PATH_ENVVAR)
    )

  def apply_opt(self, opt, val):
    ''' Apply a command line option.
    '''
    options = self.options
    if opt == '-C':
      options.calibre_path = val
    elif opt == '-K':
      db_subpaths = (
          KindleBookAssetDB.DB_FILENAME,
          joinpath(KindleTree.CONTENT_DIRNAME, KindleBookAssetDB.DB_FILENAME),
      )
      for db_subpath in db_subpaths:
        db_fspath = joinpath(val, db_subpath)
        if existspath(db_fspath):
          break
      else:
        raise GetoptError(
            "cannot find db at %s" % (" or ".join(map(repr, db_subpaths)),)
        )
      options.kindle_path = dirname(db_fspath)
    else:
      super().apply_opt(opt, val)

  @contextmanager
  def run_context(self):
    ''' Prepare the `SQLTags` around each command invocation.
    '''
    from .calibre import CalibreTree  # pylint: disable=import-outside-toplevel
    with super().run_context():
      options = self.options
      dedrm = (
          DeDRMWrapper(options.dedrm_package_path)
          if options.dedrm_package_path else None
      )

      with KindleTree(options.kindle_path) as kt:
        with CalibreTree(options.calibre_path) as cal:
          with stackattrs(options, kindle=kt, calibre=cal, dedrm=dedrm):
            yield

  def cmd_app_path(self, argv):
    ''' Usage: {cmd} [content-path]
          Report or set the content path for the Kindle application.
    '''
    if not argv:
      print(kindle_content_path())
      return 0
    content_path = self.poparg(
        argv,
        lambda arg: arg,
        "content-path",
        lambda path: path == 'DEFAULT' or isdirpath(path),
        "content-path should be DEFAULT or an existing directory",
    )
    if argv:
      raise GetoptError("extra arguments: %r" % (argv,))
    if content_path == 'DEFAULT':
      content_path = kindle_content_path_default()
    defaults = OSXDomainDefaults(KINDLE_APP_OSX_DEFAULTS_DOMAIN)
    defaults[KINDLE_APP_OSX_DEFAULTS_CONTENT_PATH_SETTING] = content_path
    return 0

  def cmd_dbshell(self, argv):
    ''' Usage: {cmd}
          Start an interactive database prompt.
    '''
    if argv:
      raise GetoptError("extra arguments: %r" % (argv,))
    return self.options.kindle.dbshell()

  # pylint: disable=too-many-locals
  def cmd_export(self, argv):
    ''' Usage: {cmd} [-fnqv] [ASINs...]
          Export AZW files to Calibre library.
          -f    Force: replace the AZW3 format if already present.
          -n    No action, recite planned actions.
          -q    Quiet: report only warnings.
          -v    Verbose: report more information about actions and inaction.
          ASINs Optional ASIN identifiers to export.
                The default is to export all books with no "calibre.dbid" fstag.
    '''
    options = self.options
    kindle = options.kindle
    calibre = options.calibre
    runstate = options.runstate
    self.popopts(argv, options, f='force', n='-doit', q='quiet', v='verbose')
    dedrm = options.dedrm
    doit = options.doit
    force = options.force
    quiet = options.quiet
    verbose = options.verbose
    asins = argv or sorted(kindle.asins())
    xit = 0
    quiet or print("export", kindle.shortpath, "=>", calibre.shortpath)
    for asin in progressbar(asins, f"export to {calibre}"):
      if runstate.cancelled:
        break
      with Pfx(asin):
        kbook = kindle.by_asin(asin)
        try:
          kbook.export_to_calibre(
              calibre,
              dedrm=dedrm,
              doit=doit,
              force=force,
              replace_format=force,
              quiet=quiet,
              verbose=verbose,
          )
        except ValueError as e:
          warning("export failed: %s", e)
          xit = 1
    if runstate.cancelled:
      xit = 1
    return xit

  def cmd_import_tags(self, argv):
    ''' Usage: {cmd} [-nqv] [ASINs...]
          Import Calibre book information into the fstags for a Kindle book.
          This will support doing searches based on stuff like
          titles which are, naturally, not presented in the Kindle
          metadata db.
    '''
    options = self.options
    kindle = options.kindle
    calibre = options.calibre
    runstate = options.runstate
    self.popopts(argv, options, n='-doit', q='quiet', v='verbose')
    doit = options.doit
    quiet = options.quiet
    verbose = options.verbose
    asins = argv or sorted(kindle.asins())
    xit = 0
    for asin in progressbar(asins, f"import metadata from {calibre}"):
      if runstate.cancelled:
        break
      with Pfx(asin):
        kbook = kindle.by_asin(asin)
        cbooks = list(calibre.by_asin(asin))
        if not cbooks:
          # pylint: disable=expression-not-assigned
          verbose and print("asin %s: no Calibre books" % (asin,))
          continue
        cbook = cbooks[0]
        if len(cbooks) > 1:
          # pylint: disable=expression-not-assigned
          quiet or print(
              f"asin {asin}: multiple Calibre books, dbids {[cb.dbid for cb in cbooks]!r}; choosing {cbook}"
          )
        ktags = kbook.tags
        ctags = ktags.subtags('calibre')
        import_tags = dict(
            title=cbook.title,
            authors=sorted([author.name for author in cbook.authors]),
            dbfspath=calibre.fspath,
            dbid=cbook.dbid,
            identifiers=cbook.identifiers,
            tags=cbook.tags,
        )
        first_update = True
        for tag_name, tag_value in sorted(import_tags.items()):
          with Pfx("%s=%r", tag_name, tag_value):
            old_value = ctags.get(tag_name)
            if old_value != tag_value:
              if not quiet:
                if first_update:
                  print(f"{asin}: update from {cbook}")
                  first_update = False
                if old_value:
                  print(
                      f"  calibre.{tag_name}={tag_value!r}, was {old_value!r}"
                  )
                else:
                  print(f"  calibre.{tag_name}={tag_value!r}")
              if doit:
                ctags[tag_name] = tag_value
    if runstate.cancelled:
      xit = 1
    return xit

  def cmd_info(self, argv):
    ''' Usage: {cmd}
          Report basic information.
    '''
    if argv:
      raise GetoptError("extra arguments: %r" % (argv,))
    print("kindle", self.options.kindle.shortpath)
    print("calibre", self.options.calibre.shortpath)

  def cmd_ls(self, argv):
    ''' Usage: {cmd} [-l]
          List the contents of the library.
          -l  Long mode.
    '''
    options = self.options
    kindle = options.kindle
    options.longmode = False
    self.popopts(argv, options, l='longmode')
    longmode = options.longmode
    if argv:
      raise GetoptError("extra arguments: %r" % (argv,))
    print(kindle.fspath)
    for subdir_name, kbook in kindle.items():
      line1 = [subdir_name]
      title = kbook.tags.auto.calibre.title
      if title:
        line1.append(title)
      authors = kbook.tags.auto.calibre.authors
      if authors:
        line1.append(','.join(authors))
      print(*line1)
      if longmode:
        if kbook.type != 'kindle.ebook':
          print("  type =", kbook.type)
        if kbook.revision is not None:
          print("  revision =", kbook.revision)
        if kbook.sampling:
          print("  sampling =", kbook.sampling)
        for tag in sorted(kbook.tags.as_tags()):
          if tag.name not in ('calibre.title', 'calibre.authors'):
            print(" ", tag)

if __name__ == '__main__':
  sys.exit(main(sys.argv))
